# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
jenkins-hash lookup3 algorithm implementation
"""

from threading import Lock
import struct

c_signed_short = struct.Struct(">h")


def rot(x, k):
    return (x << k) | (x >> (32 - k))


def mix(a, b, c):
    a &= 0xFFFFFFFF
    b &= 0xFFFFFFFF
    c &= 0xFFFFFFFF
    a -= c
    a &= 0xFFFFFFFF
    a ^= rot(c, 4)
    a &= 0xFFFFFFFF
    c += b
    c &= 0xFFFFFFFF
    b -= a
    b &= 0xFFFFFFFF
    b ^= rot(a, 6)
    b &= 0xFFFFFFFF
    a += c
    a &= 0xFFFFFFFF
    c -= b
    c &= 0xFFFFFFFF
    c ^= rot(b, 8)
    c &= 0xFFFFFFFF
    b += a
    b &= 0xFFFFFFFF
    a -= c
    a &= 0xFFFFFFFF
    a ^= rot(c, 16)
    a &= 0xFFFFFFFF
    c += b
    c &= 0xFFFFFFFF
    b -= a
    b &= 0xFFFFFFFF
    b ^= rot(a, 19)
    b &= 0xFFFFFFFF
    a += c
    a &= 0xFFFFFFFF
    c -= b
    c &= 0xFFFFFFFF
    c ^= rot(b, 4)
    c &= 0xFFFFFFFF
    b += a
    b &= 0xFFFFFFFF
    return a, b, c


def final(a, b, c):
    a &= 0xFFFFFFFF
    b &= 0xFFFFFFFF
    c &= 0xFFFFFFFF
    c ^= b
    c &= 0xFFFFFFFF
    c -= rot(b, 14)
    c &= 0xFFFFFFFF
    a ^= c
    a &= 0xFFFFFFFF
    a -= rot(c, 11)
    a &= 0xFFFFFFFF
    b ^= a
    b &= 0xFFFFFFFF
    b -= rot(a, 25)
    b &= 0xFFFFFFFF
    c ^= b
    c &= 0xFFFFFFFF
    c -= rot(b, 16)
    c &= 0xFFFFFFFF
    a ^= c
    a &= 0xFFFFFFFF
    a -= rot(c, 4)
    a &= 0xFFFFFFFF
    b ^= a
    b &= 0xFFFFFFFF
    b -= rot(a, 14)
    b &= 0xFFFFFFFF
    c ^= b
    c &= 0xFFFFFFFF
    c -= rot(b, 24)
    c &= 0xFFFFFFFF
    return a, b, c


def compute_hash(data, init_val=0, init_val2=0):
    # pylint: disable=too-many-statements,docstring-missing-param,docstring-missing-rtype,docstring-missing-return
    """
    implementation by:
    https://stackoverflow.com/questions/3279615/python-implementation-of-jenkins-hash
    """
    length = lenpos = len(data)

    a = b = c = 0xDEADBEEF + length + init_val

    c += init_val2
    c &= 0xFFFFFFFF

    p = 0  # string offset
    while lenpos > 12:
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
        a &= 0xFFFFFFFF
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        b &= 0xFFFFFFFF
        c += (
            ord(data[p + 8])
            + (ord(data[p + 9]) << 8)
            + (ord(data[p + 10]) << 16)
            + (ord(data[p + 11]) << 24)
        )
        c &= 0xFFFFFFFF
        a, b, c = mix(a, b, c)
        p += 12
        lenpos -= 12

    if lenpos == 12:
        c += (
            ord(data[p + 8])
            + (ord(data[p + 9]) << 8)
            + (ord(data[p + 10]) << 16)
            + (ord(data[p + 11]) << 24)
        )
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 11:
        c += ord(data[p + 8]) + (ord(data[p + 9]) << 8) + (ord(data[p + 10]) << 16)
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 10:
        c += ord(data[p + 8]) + (ord(data[p + 9]) << 8)
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 9:
        c += ord(data[p + 8])
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 8:
        b += (
            ord(data[p + 4])
            + (ord(data[p + 5]) << 8)
            + (ord(data[p + 6]) << 16)
            + (ord(data[p + 7]) << 24)
        )
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 7:
        b += ord(data[p + 4]) + (ord(data[p + 5]) << 8) + (ord(data[p + 6]) << 16)
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 6:
        b += (ord(data[p + 5]) << 8) + ord(data[p + 4])
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 5:
        b += ord(data[p + 4])
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 4:
        a += (
            ord(data[p + 0])
            + (ord(data[p + 1]) << 8)
            + (ord(data[p + 2]) << 16)
            + (ord(data[p + 3]) << 24)
        )
    if lenpos == 3:
        a += ord(data[p + 0]) + (ord(data[p + 1]) << 8) + (ord(data[p + 2]) << 16)
    if lenpos == 2:
        a += ord(data[p + 0]) + (ord(data[p + 1]) << 8)
    if lenpos == 1:
        a += ord(data[p + 0])

    a &= 0xFFFFFFFF
    b &= 0xFFFFFFFF
    c &= 0xFFFFFFFF
    if lenpos == 0:
        return c, b

    a, b, c = final(a, b, c)

    return c, b


def generate_hash_code(partition_key):
    if not partition_key:
        return 0

    hash_tuple = compute_hash(partition_key, 0, 0)
    hash_value = (hash_tuple[0] ^ hash_tuple[1]) & 0xFFFF
    return c_signed_short.unpack(struct.pack(">H", hash_value))[0]


class PartitionResolver:
    def __init__(self, partitions):
        self._idx = -1
        self._partitions = partitions
        self._partitions_cnt = len(self._partitions)
        self._lock = Lock()

    def get_next_partition_id(self):
        """
        round-robin partition assignment

        :return: The next partition id.
        :rtype: str
        """
        with self._lock:
            self._idx += 1
            self._idx %= self._partitions_cnt
            return self._partitions[self._idx]

    def get_partition_id_by_partition_key(self, partition_key):
        hash_code = generate_hash_code(partition_key)
        return self._partitions[abs(hash_code % self._partitions_cnt)]



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show async sending and receiving events behind a proxy.
"""
from typing import Union, Dict
import os
import asyncio
from azure.eventhub.aio import EventHubConsumerClient, EventHubProducerClient
from azure.eventhub import EventData

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']

HTTP_PROXY: Dict[str, Union[str, int]] = {
    'proxy_hostname': '127.0.0.1',  # proxy hostname.
    'proxy_port': 3128,  # proxy port.
    'username': 'admin',  # username used for proxy authentication if needed.
    'password': '123456'  # password used for proxy authentication if needed.
}


async def on_event(partition_context, event):
    # Put your code here.
    print("received event from partition: {}.".format(partition_context.partition_id))
    print(event)


async def main():
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
        http_proxy=HTTP_PROXY
    )

    producer_client = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME,
        http_proxy=HTTP_PROXY
    )

    async with producer_client:
        event_data_batch = await producer_client.create_batch(max_size_in_bytes=10000)
        while True:
            try:
                event_data_batch.add(EventData('Message inside EventBatchData'))
            except ValueError:
                # EventDataBatch object reaches max_size.
                # New EventDataBatch object can be created here to send more data.
                break
        await producer_client.send_batch(event_data_batch)
        print('Finished sending.')

    async with consumer_client:
        await consumer_client.receive(on_event=on_event)
        print('Finished receiving.')


asyncio.run(main())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
import uuid
import asyncio
import logging
from collections import deque
from typing import TYPE_CHECKING, Callable, Awaitable, Dict, Optional, Union, List
from functools import partial

from ._client_base_async import ConsumerProducerMixin
from ._async_utils import get_dict_with_loop_if_needed
from .._common import EventData
from .._utils import create_properties, event_position_selector
from .._constants import EPOCH_SYMBOL, TIMEOUT_SYMBOL, RECEIVER_RUNTIME_METRIC_SYMBOL

if TYPE_CHECKING:
    from typing import Deque

    try:
        from uamqp import (  # pylint: disable=unused-import
            ReceiveClientAsync as uamqp_ReceiveClientAsync,
            Message as uamqp_Message,
        )
        from uamqp.types import AMQPType as uamqp_AMQPType
        from uamqp.authentication import JWTTokenAsync as uamqp_JWTTokenAsync
    except ImportError:
        uamqp_Message = None
        uamqp_ReceiveClientAsync = None
        uamqp_AMQPType = None
        uamqp_JWTTokenAsync = None

    from .._pyamqp.aio._authentication_async import JWTTokenAuthAsync
    from .._pyamqp.aio._client_async import ReceiveClientAsync
    from .._pyamqp import types

    from ._consumer_client_async import EventHubConsumerClient

_LOGGER = logging.getLogger(__name__)


class EventHubConsumer(
    ConsumerProducerMixin
):  # pylint:disable=too-many-instance-attributes
    """
    A consumer responsible for reading EventData from a specific Event Hub
    partition and as a member of a specific consumer group.

    A consumer may be exclusive, which asserts ownership over the partition for the consumer
    group to ensure that only one consumer from that group is reading the from the partition.
    These exclusive consumers are sometimes referred to as "Epoch Consumers."

    A consumer may also be non-exclusive, allowing multiple consumers from the same consumer
    group to be actively reading events from the partition.  These non-exclusive consumers are
    sometimes referred to as "Non-Epoch Consumers."

    Please use the method `_create_consumer` on `EventHubClient` for creating `EventHubConsumer`.

    :param client: The parent EventHubConsumerClient.
    :type client: ~azure.eventhub.aio.EventHubConsumerClient
    :param source: The source EventHub from which to receive events.
    :type source: ~uamqp.address.Source
    :keyword event_position: The position from which to start receiving.
    :paramtype event_position: int, str, datetime.datetime
    :keyword int prefetch: The number of events to prefetch from the service
        for processing. Default is 300.
    :keyword int owner_level: The priority of the exclusive consumer. An exclusive
        consumer will be created if owner_level is set.
    :keyword bool track_last_enqueued_event_properties: Indicates whether or not the consumer should request information
        on the last enqueued event on its associated partition, and track that information as events are received.
        When information about the partition's last enqueued event is being tracked, each event received from the
        Event Hubs service will carry metadata about the partition. This results in a small amount of additional
        network bandwidth consumption that is generally a favorable trade-off when considered against periodically
        making requests for partition properties using the Event Hub client.
        It is set to `False` by default.
    """

    def __init__(self, client: "EventHubConsumerClient", source: str, **kwargs) -> None:
        super().__init__()
        event_position = kwargs.get("event_position", None)
        prefetch = kwargs.get("prefetch", 300)
        owner_level = kwargs.get("owner_level", None)
        keep_alive = kwargs.get("keep_alive", None)
        auto_reconnect = kwargs.get("auto_reconnect", True)
        track_last_enqueued_event_properties = kwargs.get(
            "track_last_enqueued_event_properties", False
        )
        idle_timeout = kwargs.get("idle_timeout", None)

        self.running = False
        self.closed = False

        self._amqp_transport = kwargs.pop("amqp_transport")
        self._on_event_received: Callable[
            [Union[Optional[EventData], List[EventData]]], Awaitable[None]
        ] = kwargs["on_event_received"]
        self._internal_kwargs = get_dict_with_loop_if_needed(kwargs.get("loop", None))
        self._client = client
        self._source = source
        self._offset = event_position
        self._offset_inclusive = kwargs.get("event_position_inclusive", False)
        self._prefetch = prefetch
        self._owner_level = owner_level
        self._keep_alive = keep_alive
        self._auto_reconnect = auto_reconnect
        self._retry_policy = self._amqp_transport.create_retry_policy(
            self._client._config
        )
        self._reconnect_backoff = 1
        self._timeout = 0
        self._idle_timeout = (
            (idle_timeout * self._amqp_transport.TIMEOUT_FACTOR)
            if idle_timeout
            else None
        )
        link_properties: Dict[bytes, int] = {}
        self._partition = self._source.split("/")[-1]
        self._name = f"EHReceiver-{uuid.uuid4()}-partition{self._partition}"
        if owner_level is not None:
            link_properties[EPOCH_SYMBOL] = int(owner_level)
        link_property_timeout_ms = (
            self._client._config.receive_timeout
            or self._timeout  # pylint:disable=protected-access
        ) * self._amqp_transport.TIMEOUT_FACTOR
        link_properties[TIMEOUT_SYMBOL] = int(link_property_timeout_ms)
        self._link_properties: Union[
            Dict[uamqp_AMQPType, uamqp_AMQPType], Dict[types.AMQPTypes, types.AMQPTypes]
        ] = self._amqp_transport.create_link_properties(link_properties)
        self._handler: Optional[ReceiveClientAsync] = None
        self._track_last_enqueued_event_properties = (
            track_last_enqueued_event_properties
        )
        self._message_buffer: Deque[uamqp_Message] = deque()
        self._last_received_event: Optional[EventData] = None
        self._message_buffer_lock = asyncio.Lock()
        self._last_callback_called_time = None
        self._callback_task_run = None

    def _create_handler(
        self, auth: Union[uamqp_JWTTokenAsync, JWTTokenAuthAsync]
    ) -> None:
        source = self._amqp_transport.create_source(
            self._source,
            self._offset,
            event_position_selector(self._offset, self._offset_inclusive),
        )
        desired_capabilities = (
            [RECEIVER_RUNTIME_METRIC_SYMBOL]
            if self._track_last_enqueued_event_properties
            else None
        )

        self._handler = self._amqp_transport.create_receive_client(
            config=self._client._config,  # pylint:disable=protected-access
            source=source,
            auth=auth,
            network_trace=self._client._config.network_tracing,  # pylint:disable=protected-access
            link_credit=self._prefetch,
            link_properties=self._link_properties,
            idle_timeout=self._idle_timeout,
            retry_policy=self._retry_policy,
            keep_alive_interval=self._keep_alive,
            client_name=self._name,
            properties=create_properties(
                self._client._config.user_agent,  # pylint:disable=protected-access
                amqp_transport=self._amqp_transport,
            ),
            desired_capabilities=desired_capabilities,
            streaming_receive=True,
            message_received_callback=partial(
                self._amqp_transport.message_received_async, self
            ),
        )

    async def _open_with_retry(self) -> None:
        await self._do_retryable_operation(self._open, operation_need_param=False)

    def _next_message_in_buffer(self):
        # pylint:disable=protected-access
        message = self._message_buffer.popleft()
        event_data = EventData._from_message(message)
        if self._amqp_transport.KIND == "uamqp":
            event_data._uamqp_message = message
        self._last_received_event = event_data
        return event_data

    async def receive(
        self, batch=False, max_batch_size=300, max_wait_time=None
    ) -> None:
        await self._amqp_transport.receive_messages_async(
            self, batch, max_batch_size, max_wait_time
        )



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------

"""Fake implementation of AbstractSpan for tests. Copied from `azure-core/tests`."""
from contextlib import contextmanager
from typing import Union, Sequence, Optional, Dict
import uuid

from azure.core.tracing import HttpSpanMixin, SpanKind


AttributeValue = Union[
    str,
    bool,
    int,
    float,
    Sequence[str],
    Sequence[bool],
    Sequence[int],
    Sequence[float],
]
Attributes = Optional[Dict[str, AttributeValue]]


class FakeSpan(HttpSpanMixin, object):
    # Keep a fake context of the current one
    CONTEXT = []

    def __init__(self, span=None, name="span", kind=None, *, links=[], **kwargs):
        # type: (Optional[Span], Optional[str]) -> None
        """
        If a span is not passed in, creates a new tracer. If the instrumentation key for Azure Exporter is given, will
        configure the azure exporter else will just create a new tracer.

        :param span: The OpenTelemetry span to wrap
        :type span: :class: OpenTelemetry.trace.Span
        :param name: The name of the OpenTelemetry span to create if a new span is needed
        :type name: str
        """
        self._span = span
        self.name = name
        self._kind = kind or SpanKind.UNSPECIFIED
        self.attributes = {}
        self.children = []
        self.status = None
        self.links = links
        self.traceparent = uuid.uuid4().hex

    def __str__(self):
        buffer = "Name: {}\n".format(self.name)
        buffer += "Children:\n"
        subchildren = "\n".join(str(child) for child in self.children)
        buffer += "\n".join("\t{}".format(line) for line in subchildren.splitlines())
        return buffer

    @property
    def span_instance(self):
        # type: () -> Span
        """
        :return: The OpenTelemetry span that is being wrapped.
        """
        return self._span

    def span(self, name="span", kind=None, links=[]):
        # type: (Optional[str]) -> OpenCensusSpan
        """
        Create a child span for the current span and append it to the child spans list in the span instance.
        :param name: Name of the child span
        :type name: str
        :return: The OpenCensusSpan that is wrapping the child span instance
        """
        return self.__class__(name=name, kind=kind, links=links)

    @property
    def kind(self):
        # type: () -> Optional[SpanKind]
        """Get the span kind of this span."""
        return self._kind

    @kind.setter
    def kind(self, value):
        # type: (SpanKind) -> None
        """Set the span kind of this span."""
        self._kind = value

    def __enter__(self):
        """Start a span."""
        if self.CONTEXT:
            self.CONTEXT[-1].children.append(self)
        self.CONTEXT.append(self)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Finish a span."""
        if exception_value:
            self.status = exception_value.args[0]
        if self.CONTEXT:
            self.CONTEXT.pop()

    def start(self):
        # type: () -> None
        """Set the start time for a span."""
        pass

    def finish(self):
        # type: () -> None
        """Set the end time for a span."""
        self.CONTEXT.pop()

    def to_header(self):
        # type: () -> Dict[str, str]
        """
        Returns a dictionary with the header labels and values.
        :return: A key value pair dictionary
        """
        return {'traceparent': self.traceparent}

    def add_attribute(self, key, value):
        # type: (str, Union[str, int]) -> None
        """
        Add attribute (key value pair) to the current span.

        :param key: The key of the key value pair
        :type key: str
        :param value: The value of the key value pair
        :type value: str
        """
        self.attributes[key] = value

    def get_trace_parent(self):
        """Return traceparent string as defined in W3C trace context specification.

        Example:
        Value = 00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01
        base16(version) = 00
        base16(trace-id) = 4bf92f3577b34da6a3ce929d0e0e4736
        base16(parent-id) = 00f067aa0ba902b7
        base16(trace-flags) = 01  // sampled

        :return: a traceparent string
        :rtype: str
        """
        return self.to_header()['traceparent']

    @classmethod
    def link(cls, traceparent, attributes=None):
        # type: (str, Attributes) -> None
        """
        Links the context to the current tracer.

        :param traceparent: A complete traceparent
        :type traceparent: str
        """
        cls.link_from_headers({
            'traceparent': traceparent
        })

    @classmethod
    def link_from_headers(cls, headers, attributes=None):
        # type: (Dict[str, str], Attributes) -> None
        """
        Given a dictionary, extracts the context and links the context to the current tracer.

        :param headers: A key value pair dictionary
        :type headers: dict
        """
        raise NotImplementedError("This method needs to be implemented")

    @classmethod
    def get_current_span(cls):
        # type: () -> Span
        """
        Get the current span from the execution context. Return None otherwise.
        """
        return cls.CONTEXT[-1]

    @classmethod
    def get_current_tracer(cls):
        # type: () -> Tracer
        """
        Get the current tracer from the execution context. Return None otherwise.
        """
        raise NotImplementedError()

    @classmethod
    @contextmanager
    def change_context(cls, span):
        # type: (Span) -> ContextManager
        """Change the context for the life of this context manager.
        """
        try:
            cls.CONTEXT.append(span)
            yield
        finally:
            cls.CONTEXT.pop()

    @classmethod
    def set_current_span(cls, span):
        # type: (Span) -> None
        """Not supported by OpenTelemetry.
        """
        raise NotImplementedError()

    @classmethod
    def set_current_tracer(cls, tracer):
        # type: (Tracer) -> None
        """
        Set the given tracer as the current tracer in the execution context.
        :param tracer: The tracer to set the current tracer as
        :type tracer: :class: OpenTelemetry.trace.Tracer
        """
        raise NotImplementedError()

    @classmethod
    def with_current_context(cls, func):
        # type: (Callable) -> Callable
        """Passes the current spans to the new context the function will be run in.

        :param func: The function that will be run in the new context
        :return: The target the pass in instead of the function
        """
        return func



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import pytest
import time
import sys
import threading

from azure.identity import EnvironmentCredential
from azure.eventhub import (
    EventData,
    EventDataBatch)
from azure.eventhub.exceptions import (
    ConnectError,
    AuthenticationError,
    EventHubError,
    OperationTimeoutError
)
from azure.eventhub import (
    EventHubProducerClient,
    EventHubConsumerClient,
    EventHubSharedKeyCredential
)
from azure.eventhub._client_base import EventHubSASTokenCredential


@pytest.mark.liveTest
def test_send_batch_with_invalid_hostname(invalid_hostname, uamqp_transport):
    if sys.platform.startswith('darwin'):
        pytest.skip("Skipping on OSX - it keeps reporting 'Unable to set external certificates' "
                    "and blocking other tests")
    client = EventHubProducerClient.from_connection_string(invalid_hostname, uamqp_transport=uamqp_transport)
    with client:
        with pytest.raises(ConnectError):
            batch = EventDataBatch()
            batch.add(EventData("test data"))
            client.send_batch(batch)

    # test setting callback
    def on_error(events, pid, err):
        assert len(events) == 1
        assert not pid
        on_error.err = err

    on_error.err = None
    client = EventHubProducerClient.from_connection_string(invalid_hostname, on_error=on_error, uamqp_transport=uamqp_transport)
    with client:
        batch = EventDataBatch()
        batch.add(EventData("test data"))
        client.send_batch(batch)
    assert isinstance(on_error.err, ConnectError)

    on_error.err = None
    client = EventHubProducerClient.from_connection_string(invalid_hostname, on_error=on_error, uamqp_transport=uamqp_transport)
    with client:
        client.send_event(EventData("test data"))
    assert isinstance(on_error.err, ConnectError)


@pytest.mark.liveTest
def test_receive_with_invalid_hostname_sync(invalid_hostname, uamqp_transport):
    def on_event(partition_context, event):
        pass

    client = EventHubConsumerClient.from_connection_string(
        invalid_hostname, consumer_group='$default', uamqp_transport=uamqp_transport
    )
    with client:
        thread = threading.Thread(target=client.receive,
                                  args=(on_event, ))
        thread.start()
        time.sleep(2)
        assert len(client._event_processors) == 1
    thread.join()


@pytest.mark.liveTest
def test_send_batch_with_invalid_key(invalid_key, uamqp_transport):
    client = EventHubProducerClient.from_connection_string(invalid_key, uamqp_transport=uamqp_transport)
    try:
        with pytest.raises(ConnectError):
            batch = EventDataBatch()
            batch.add(EventData("test data"))
            client.send_batch(batch)
    finally:
        client.close()


@pytest.mark.liveTest
def test_send_batch_to_invalid_partitions(connection_str, uamqp_transport):
    partitions = ["XYZ", "-1", "1000", "-"]
    for p in partitions:
        client = EventHubProducerClient.from_connection_string(connection_str, uamqp_transport=uamqp_transport)
        try:
            with pytest.raises(ConnectError):
                batch = client.create_batch(partition_id=p)
                batch.add(EventData("test data"))
                client.send_batch(batch)
        finally:
            client.close()


@pytest.mark.liveTest
def test_send_batch_too_large_message(connection_str, uamqp_transport):
    if sys.platform.startswith('darwin'):
        pytest.skip("Skipping on OSX - open issue regarding message size")
    client = EventHubProducerClient.from_connection_string(connection_str, uamqp_transport=uamqp_transport)
    try:
        data = EventData(b"A" * 1100000)
        batch = client.create_batch()
        with pytest.raises(ValueError):
            batch.add(data)
    finally:
        client.close()


@pytest.mark.liveTest
def test_send_batch_null_body(connection_str, uamqp_transport):
    client = EventHubProducerClient.from_connection_string(connection_str, uamqp_transport=uamqp_transport)
    try:
        with pytest.raises(ValueError):
            data = EventData(None)
            batch = client.create_batch()
            batch.add(data)
            client.send_batch(batch)
    finally:
        client.close()


@pytest.mark.liveTest
def test_create_batch_with_invalid_hostname_sync(invalid_hostname, uamqp_transport):
    if sys.platform.startswith('darwin'):
        pytest.skip("Skipping on OSX - it keeps reporting 'Unable to set external certificates' "
                    "and blocking other tests")
    client = EventHubProducerClient.from_connection_string(invalid_hostname, uamqp_transport=uamqp_transport)
    with client:
        with pytest.raises(ConnectError):
            client.create_batch(max_size_in_bytes=300)


@pytest.mark.liveTest
def test_create_batch_with_too_large_size_sync(connection_str, uamqp_transport):
    client = EventHubProducerClient.from_connection_string(connection_str, uamqp_transport=uamqp_transport)
    with client:
        with pytest.raises(ValueError):
            client.create_batch(max_size_in_bytes=5 * 1024 * 1024)

@pytest.mark.liveTest
def test_invalid_proxy_server(connection_str, uamqp_transport):
    if sys.platform.startswith('darwin') and uamqp_transport:
        pytest.skip("Skipping on OSX - running forever and blocking other tests")
    HTTP_PROXY = {
    'proxy_hostname': 'fakeproxy',  # proxy hostname.
    'proxy_port': 3128,  # proxy port.
    }

    client = EventHubProducerClient.from_connection_string(connection_str, http_proxy=HTTP_PROXY, uamqp_transport=uamqp_transport)
    with client:
        with pytest.raises(EventHubError):
            batch = client.create_batch()

@pytest.mark.liveTest
def test_client_send_timeout(connstr_receivers, uamqp_transport):
    connection_str, receivers = connstr_receivers

    def on_success(events, pid):
        pass

    def on_error(events, pid, err):
        pass

    producer = EventHubProducerClient.from_connection_string(
        connection_str, uamqp_transport=uamqp_transport
    )

    with producer:
        with pytest.raises(OperationTimeoutError):
            producer.send_batch([EventData(b"Data")], timeout=-1)

        with pytest.raises(OperationTimeoutError):
            producer.send_event(EventData(b"Data"), timeout=-1)

    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        with pytest.raises(OperationTimeoutError):
            producer.send_batch([EventData(b"Data")], timeout=-1)

        with pytest.raises(OperationTimeoutError):
            producer.send_event(EventData(b"Data"), timeout=-1)


@pytest.mark.liveTest
def test_client_invalid_credential(live_eventhub, uamqp_transport):

    def on_event(partition_context, event):
        pass

    def on_error(partition_context, error):
        on_error.err = error

    env_credential = EnvironmentCredential()
    # Skipping on OSX - it's raising a ConnectionLostError
    if not sys.platform.startswith('darwin'):
        producer_client = EventHubProducerClient(fully_qualified_namespace="fakeeventhub.servicebus.windows.net",
                                                eventhub_name=live_eventhub['event_hub'],
                                                credential=env_credential,
                                                user_agent='customized information',
                                                retry_total=1,
                                                retry_mode='exponential',
                                                retry_backoff=0.02,
                                                uamqp_transport=uamqp_transport)
        consumer_client = EventHubConsumerClient(fully_qualified_namespace="fakeeventhub.servicebus.windows.net",
                                                eventhub_name=live_eventhub['event_hub'],
                                                credential=env_credential,
                                                user_agent='customized information',
                                                consumer_group='$Default',
                                                retry_total=1,
                                                retry_mode='exponential',
                                                retry_backoff=0.02,
                                                uamqp_transport=uamqp_transport)
        with producer_client:
            with pytest.raises(ConnectError):
                producer_client.create_batch(partition_id='0')

        on_error.err = None
        with consumer_client:
            thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                    kwargs={"starting_position": "-1", "on_error": on_error})
            thread.daemon = True
            thread.start()
            time.sleep(15)
        thread.join()
        assert isinstance(on_error.err, ConnectError)

    producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name='fakehub',
                                             credential=env_credential,
                                             uamqp_transport=uamqp_transport)

    consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name='fakehub',
                                             credential=env_credential,
                                             consumer_group='$Default',
                                             retry_total=0,
                                             uamqp_transport=uamqp_transport)

    with producer_client:
        with pytest.raises(ConnectError):
            producer_client.create_batch(partition_id='0')

    on_error.err = None
    with consumer_client:
        thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                  kwargs={"starting_position": "-1", "on_error": on_error})
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential(live_eventhub['key_name'], live_eventhub['access_key'])
    auth_uri = "sb://{}/{}".format(live_eventhub['hostname'], live_eventhub['event_hub'])
    token = credential.get_token(auth_uri).token
    producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=EventHubSASTokenCredential(token, time.time() + 5),
                                             uamqp_transport=uamqp_transport)
    time.sleep(10)
    # expired credential
    with producer_client:
        with pytest.raises(AuthenticationError):
            producer_client.create_batch(partition_id='0')

    consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                            eventhub_name=live_eventhub['event_hub'],
                                            credential=EventHubSASTokenCredential(token, time.time() + 7),
                                            consumer_group='$Default',
                                            retry_total=0,
                                            uamqp_transport=uamqp_transport)
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                kwargs={"starting_position": "-1", "on_error": on_error})
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()

    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential('fakekey', live_eventhub['access_key'])
    producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             uamqp_transport=uamqp_transport)
    with producer_client:
        with pytest.raises(AuthenticationError):
            producer_client.create_batch(partition_id='0')

    consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             consumer_group='$Default',
                                             retry_total=0,
                                             uamqp_transport=uamqp_transport)
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                  kwargs={"starting_position": "-1", "on_error": on_error})
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    credential = EventHubSharedKeyCredential(live_eventhub['key_name'], 'fakekey')
    producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             uamqp_transport=uamqp_transport)

    with producer_client:
        errors = (AuthenticationError)
        # TODO: flaky TimeoutError during connect for China region
        if 'servicebus.windows.net' not in live_eventhub['hostname']:
            errors = (AuthenticationError, ConnectError)
        with pytest.raises(errors):
            producer_client.create_batch(partition_id='0')

    consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             consumer_group='$Default',
                                             retry_total=0,
                                             uamqp_transport=uamqp_transport)
    on_error.err = None
    with consumer_client:
        thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                  kwargs={"starting_position": "-1", "on_error": on_error})
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()
    assert isinstance(on_error.err, AuthenticationError)

    producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=env_credential,
                                             connection_verify="cacert.pem",
                                             uamqp_transport=uamqp_transport)

    # TODO: this seems like a bug from uamqp, should be ConnectError?
    with producer_client:
        with pytest.raises(EventHubError):
            producer_client.create_batch(partition_id='0')

    consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             consumer_group='$Default',
                                             credential=env_credential,
                                             retry_total=0,
                                             connection_verify="cacert.pem",
                                             uamqp_transport=uamqp_transport)
    with consumer_client:
        thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                  kwargs={"starting_position": "-1", "on_error": on_error})
        thread.daemon = True
        thread.start()
        time.sleep(15)
    thread.join()

    # TODO: this seems like a bug from uamqp, should be ConnectError?
    assert isinstance(on_error.err, FileNotFoundError)

    # Skipping on OSX - it's raising a ConnectionLostError
    if not sys.platform.startswith('darwin'):
        producer_client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                                eventhub_name=live_eventhub['event_hub'],
                                                credential=env_credential,
                                                custom_endpoint_address="fakeaddr",
                                                uamqp_transport=uamqp_transport)

        with producer_client:
            with pytest.raises(AuthenticationError):
                producer_client.create_batch(partition_id='0')

        consumer_client = EventHubConsumerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                                eventhub_name=live_eventhub['event_hub'],
                                                consumer_group='$Default',
                                                credential=env_credential,
                                                retry_total=0,
                                                custom_endpoint_address="fakeaddr",
                                                uamqp_transport=uamqp_transport)
        with consumer_client:
            thread = threading.Thread(target=consumer_client.receive, args=(on_event,),
                                    kwargs={"starting_position": "-1", "on_error": on_error})
            thread.daemon = True
            thread.start()
            time.sleep(15)
        thread.join()

        assert isinstance(on_error.err, AuthenticationError)



"""Platform compatibility."""
# pylint: skip-file

from __future__ import absolute_import, unicode_literals

from typing import Tuple, cast
import platform
import re
import struct
import sys

# Jython does not have this attribute
try:
    from socket import SOL_TCP
except ImportError:  # pragma: no cover
    from socket import IPPROTO_TCP as SOL_TCP  # noqa


RE_NUM = re.compile(r'(\d+).+')


def _linux_version_to_tuple(s):
    # type: (str) -> Tuple[int, int, int]
    return cast(Tuple[int, int, int], tuple(map(_versionatom, s.split('.')[:3])))


def _versionatom(s):
    # type: (str) -> int
    if s.isdigit():
        return int(s)
    match = RE_NUM.match(s)
    return int(match.groups()[0]) if match else 0


# available socket options for TCP level
KNOWN_TCP_OPTS = {
    'TCP_CORK', 'TCP_DEFER_ACCEPT', 'TCP_KEEPCNT',
    'TCP_KEEPIDLE', 'TCP_KEEPINTVL', 'TCP_LINGER2',
    'TCP_MAXSEG', 'TCP_NODELAY', 'TCP_QUICKACK',
    'TCP_SYNCNT', 'TCP_USER_TIMEOUT', 'TCP_WINDOW_CLAMP',
}

LINUX_VERSION = None
if sys.platform.startswith('linux'):
    LINUX_VERSION = _linux_version_to_tuple(platform.release())
    if LINUX_VERSION < (2, 6, 37):
        KNOWN_TCP_OPTS.remove('TCP_USER_TIMEOUT')

    # Windows Subsystem for Linux is an edge-case: the Python socket library
    # returns most TCP_* enums, but they aren't actually supported
    if platform.release().endswith("Microsoft"):
        KNOWN_TCP_OPTS = {'TCP_NODELAY', 'TCP_KEEPIDLE', 'TCP_KEEPINTVL',
                          'TCP_KEEPCNT'}

elif sys.platform.startswith('darwin'):
    KNOWN_TCP_OPTS.remove('TCP_USER_TIMEOUT')

elif 'bsd' in sys.platform:
    KNOWN_TCP_OPTS.remove('TCP_USER_TIMEOUT')

# According to MSDN Windows platforms support getsockopt(TCP_MAXSSEG) but not
# setsockopt(TCP_MAXSEG) on IPPROTO_TCP sockets.
elif sys.platform.startswith('win'):
    KNOWN_TCP_OPTS = {'TCP_NODELAY'}

elif sys.platform.startswith('cygwin'):
    KNOWN_TCP_OPTS = {'TCP_NODELAY'}

# illumos does not allow to set the TCP_MAXSEG socket option,
# even if the Oracle documentation says otherwise.
elif sys.platform.startswith('sunos'):
    KNOWN_TCP_OPTS.remove('TCP_MAXSEG')

# aix does not allow to set the TCP_MAXSEG
# or the TCP_USER_TIMEOUT socket options.
elif sys.platform.startswith('aix'):
    KNOWN_TCP_OPTS.remove('TCP_MAXSEG')
    KNOWN_TCP_OPTS.remove('TCP_USER_TIMEOUT')

if sys.version_info < (2, 7, 7):  # pragma: no cover
    import functools

    def _to_bytes_arg(fun):
        @functools.wraps(fun)
        def _inner(s, *args, **kwargs):
            return fun(s.encode(), *args, **kwargs)
        return _inner

    pack = _to_bytes_arg(struct.pack)
    pack_into = _to_bytes_arg(struct.pack_into)
    unpack = _to_bytes_arg(struct.unpack)
    unpack_from = _to_bytes_arg(struct.unpack_from)
else:
    pack = struct.pack
    pack_into = struct.pack_into
    unpack = struct.unpack
    unpack_from = struct.unpack_from

__all__ = [
    'LINUX_VERSION',
    'SOL_TCP',
    'KNOWN_TCP_OPTS',
    'pack',
    'pack_into',
    'unpack',
    'unpack_from',
]



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#-------------------------------------------------------------------------

import time
from collections import namedtuple
from functools import partial

from .sasl import SASLAnonymousCredential, SASLPlainCredential
from .utils import generate_sas_token

from .constants import (
    AUTH_DEFAULT_EXPIRATION_SECONDS,
    TOKEN_TYPE_JWT,
    TOKEN_TYPE_SASTOKEN,
    AUTH_TYPE_CBS,
    AUTH_TYPE_SASL_PLAIN
)

AccessToken = namedtuple("AccessToken", ["token", "expires_on"])


def _generate_sas_access_token(auth_uri, sas_name, sas_key, expiry_in=AUTH_DEFAULT_EXPIRATION_SECONDS):
    expires_on = int(time.time() + expiry_in)
    token = generate_sas_token(auth_uri, sas_name, sas_key, expires_on)
    return AccessToken(
        token,
        expires_on
    )


class SASLPlainAuth(object):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    auth_type = AUTH_TYPE_SASL_PLAIN

    def __init__(self, authcid, passwd, authzid=None):
        self.sasl = SASLPlainCredential(authcid, passwd, authzid)


class _CBSAuth(object):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    auth_type = AUTH_TYPE_CBS

    def __init__(
        self,
        uri,
        audience,
        token_type,
        get_token,
        **kwargs
    ):
        """
        CBS authentication using JWT tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param get_token: The callback function used for getting and refreshing
         tokens. It should return a valid jwt token each time it is called.
        :type get_token: callable object
        :param token_type: The type field of the token request.
         Default value is `"jwt"`.
        :type token_type: str

        """
        self.sasl = SASLAnonymousCredential()
        self.uri = uri
        self.audience = audience
        self.token_type = token_type
        self.get_token = get_token
        self.expires_in = kwargs.pop("expires_in", AUTH_DEFAULT_EXPIRATION_SECONDS)
        self.expires_on = kwargs.pop("expires_on", None)

    @staticmethod
    def _set_expiry(expires_in, expires_on):
        if not expires_on and not expires_in:
            raise ValueError("Must specify either 'expires_on' or 'expires_in'.")
        if not expires_on:
            expires_on = time.time() + expires_in
        else:
            expires_in = expires_on - time.time()
            if expires_in < 1:
                raise ValueError("Token has already expired.")
        return expires_in, expires_on


class JWTTokenAuth(_CBSAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    def __init__(
        self,
        uri,
        audience,
        get_token,
        **kwargs
    ):
        """
        CBS authentication using JWT tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param get_token: The callback function used for getting and refreshing
         tokens. It should return a valid jwt token each time it is called.
        :type get_token: callable object
        :param token_type: The type field of the token request.
         Default value is `"jwt"`.
        :type token_type: str

        """
        super(JWTTokenAuth, self).__init__(uri, audience, kwargs.pop("token_type", TOKEN_TYPE_JWT), get_token)
        self.get_token = get_token


class SASTokenAuth(_CBSAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    def __init__(
        self,
        uri,
        audience,
        username,
        password,
        **kwargs
    ):
        """
        CBS authentication using SAS tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param username: The SAS token username, also referred to as the key
         name or policy name. This can optionally be encoded into the URI.
        :type username: str
        :param password: The SAS token password, also referred to as the key.
         This can optionally be encoded into the URI.
        :type password: str
        :param expires_in: The total remaining seconds until the token
         expires.
        :type expires_in: int
        :param expires_on: The timestamp at which the SAS token will expire
         formatted as seconds since epoch.
        :type expires_on: float
        :param token_type: The type field of the token request.
         Default value is `"servicebus.windows.net:sastoken"`.
        :type token_type: str

        """
        self.username = username
        self.password = password
        expires_in = kwargs.pop("expires_in", AUTH_DEFAULT_EXPIRATION_SECONDS)
        expires_on = kwargs.pop("expires_on", None)
        expires_in, expires_on = self._set_expiry(expires_in, expires_on)
        self.get_token = partial(_generate_sas_access_token, uri, username, password, expires_in)
        super(SASTokenAuth, self).__init__(
            uri,
            audience,
            kwargs.pop("token_type", TOKEN_TYPE_SASTOKEN),
            self.get_token,
            expires_in=expires_in,
            expires_on=expires_on
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

class EventHubError(Exception):
    """Represents an error occurred in the client.

    :ivar message: The error message.
    :vartype message: str
    :ivar error: The error condition, if available.
    :vartype error: str
    :ivar details: The error details, if included in the
     service response.
    :vartype details: dict[str, str]
    """

    def __init__(self, message, details=None):
        self.error = None
        self.message = message
        self.details = details
        if details and isinstance(details, Exception):
            try:
                condition = details.condition.value.decode("UTF-8")
            except AttributeError:
                try:
                    condition = details.condition.decode("UTF-8")
                except AttributeError:
                    condition = None
            if condition:
                _, _, self.error = condition.partition(":")
                self.message += "\nError: {}".format(self.error)
            try:
                self._parse_error(details.description)
                for detail in self.details:
                    self.message += "\n{}".format(detail)
            except:  # pylint: disable=bare-except
                self.message += "\n{}".format(details)
        super(EventHubError, self).__init__(self.message)

    def _parse_error(self, error_list):
        details = []
        self.message = (
            error_list
            if isinstance(error_list, str)
            else error_list.decode("UTF-8")
        )
        details_index = self.message.find(" Reference:")
        if details_index >= 0:
            details_msg = self.message[details_index + 1 :]
            self.message = self.message[0:details_index]

            tracking_index = details_msg.index(", TrackingId:")
            system_index = details_msg.index(", SystemTracker:")
            timestamp_index = details_msg.index(", Timestamp:")
            details.append(details_msg[:tracking_index])
            details.append(details_msg[tracking_index + 2 : system_index])
            details.append(details_msg[system_index + 2 : timestamp_index])
            details.append(details_msg[timestamp_index + 2 :])
            self.details = details


class ClientClosedError(EventHubError):
    """The Client has been closed and is unable to process further events."""


class ConnectionLostError(EventHubError):
    """Connection to the Event Hub is lost.

    In most cases the client will automatically retry on this error."""


class ConnectError(EventHubError):
    """Failed to connect to the Event Hubs service."""


class AuthenticationError(ConnectError):
    """Failed to connect to the Event Hubs service because of an authentication issue."""


class EventDataError(EventHubError):
    """Client prevented problematic event data from being sent."""


class EventDataSendError(EventHubError):
    """Service returned an error while an event data is being sent."""


class OperationTimeoutError(EventHubError):
    """Operation timed out."""


class OwnershipLostError(Exception):
    """Raised when `update_checkpoint` detects the ownership to a partition has been lost."""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import struct
import uuid
import logging
import time
import asyncio

from .._encode import encode_payload
from ._link_async import Link
from ..constants import SessionTransferState, LinkDeliverySettleReason, LinkState, Role, SenderSettleMode, SessionState
from ..error import AMQPLinkError, ErrorCondition, MessageException

_LOGGER = logging.getLogger(__name__)


class PendingDelivery(object):
    def __init__(self, **kwargs):
        self.message = kwargs.get("message")
        self.sent = False
        self.frame = None
        self.on_delivery_settled = kwargs.get("on_delivery_settled")
        self.start = time.time()
        self.transfer_state = None
        self.timeout = kwargs.get("timeout")
        self.settled = kwargs.get("settled", False)
        self._network_trace_params = kwargs.get('network_trace_params')

    async def on_settled(self, reason, state):
        if self.on_delivery_settled and not self.settled:
            try:
                await self.on_delivery_settled(reason, state)
            except Exception as e:  # pylint:disable=broad-except
                _LOGGER.warning(
                    "Message 'on_send_complete' callback failed: %r",
                    e,
                    extra=self._network_trace_params
                )
        self.settled = True


class SenderLink(Link):
    def __init__(self, session, handle, target_address, **kwargs):
        name = kwargs.pop("name", None) or str(uuid.uuid4())
        role = Role.Sender
        if "source_address" not in kwargs:
            kwargs["source_address"] = "sender-link-{}".format(name)
        super(SenderLink, self).__init__(session, handle, name, role, target_address=target_address, **kwargs)
        self._pending_deliveries = []

    @classmethod
    def from_incoming_frame(cls, session, handle, frame):
        # TODO: Assuming we establish all links for now...
        # check link_create_from_endpoint in C lib
        raise NotImplementedError("Pending")

    # In theory we should not need to purge pending deliveries on attach/dettach - as a link should
    # be resume-able, however this is not yet supported.
    async def _incoming_attach(self, frame):
        try:
            await super(SenderLink, self)._incoming_attach(frame)
        except AMQPLinkError:
            await self._remove_pending_deliveries()
            raise
        self.current_link_credit = self.link_credit
        await self._outgoing_flow()
        await self.update_pending_deliveries()

    async def _incoming_detach(self, frame):
        await super(SenderLink, self)._incoming_detach(frame)
        await self._remove_pending_deliveries()

    async def _incoming_flow(self, frame):
        rcv_link_credit = frame[6]  # link_credit
        rcv_delivery_count = frame[5]  # delivery_count
        if frame[4] is not None:  # handle
            if rcv_link_credit is None or rcv_delivery_count is None:
                _LOGGER.info(
                    "Unable to get link-credit or delivery-count from incoming ATTACH. Detaching link.",
                    extra=self.network_trace_params
                )
                await self._remove_pending_deliveries()
                await self._set_state(LinkState.DETACHED)  # TODO: Send detach now?
            else:
                self.current_link_credit = rcv_delivery_count + rcv_link_credit - self.delivery_count
        await self.update_pending_deliveries()

    async def _outgoing_transfer(self, delivery):
        output = bytearray()
        encode_payload(output, delivery.message)
        delivery_count = self.delivery_count + 1
        delivery.frame = {
            "handle": self.handle,
            "delivery_tag": struct.pack(">I", abs(delivery_count)),
            "message_format": delivery.message._code,  # pylint:disable=protected-access
            "settled": delivery.settled,
            "more": False,
            "rcv_settle_mode": None,
            "state": None,
            "resume": None,
            "aborted": None,
            "batchable": None,
            "payload": output,
        }
        await self._session._outgoing_transfer(  # pylint:disable=protected-access
            delivery,
            self.network_trace_params if self.network_trace else None
        )
        sent_and_settled = False
        if delivery.transfer_state == SessionTransferState.OKAY:
            self.delivery_count = delivery_count
            self.current_link_credit -= 1
            delivery.sent = True
            if delivery.settled:
                await delivery.on_settled(LinkDeliverySettleReason.SETTLED, None)
                sent_and_settled = True
        # elif delivery.transfer_state == SessionTransferState.ERROR:
        # TODO: Session wasn't mapped yet - re-adding to the outgoing delivery queue?
        return sent_and_settled

    async def _incoming_disposition(self, frame):
        if not frame[3]:  # settled
            return
        range_end = (frame[2] or frame[1]) + 1  # first or last
        settled_ids = list(range(frame[1], range_end))
        unsettled = []
        for delivery in self._pending_deliveries:
            if delivery.sent and delivery.frame["delivery_id"] in settled_ids:
                await delivery.on_settled(LinkDeliverySettleReason.DISPOSITION_RECEIVED, frame[4])  # state
                continue
            unsettled.append(delivery)
        self._pending_deliveries = unsettled

    async def _remove_pending_deliveries(self):
        futures = []
        for delivery in self._pending_deliveries:
            futures.append(asyncio.ensure_future(delivery.on_settled(LinkDeliverySettleReason.NOT_DELIVERED, None)))
        await asyncio.gather(*futures)
        self._pending_deliveries = []

    async def _on_session_state_change(self):
        if self._session.state == SessionState.DISCARDING:
            await self._remove_pending_deliveries()
        await super()._on_session_state_change()

    async def update_pending_deliveries(self):
        if self.current_link_credit <= 0:
            self.current_link_credit = self.link_credit
            await self._outgoing_flow()
        now = time.time()
        pending = []
        for delivery in self._pending_deliveries:
            if delivery.timeout and (now - delivery.start) >= delivery.timeout:
                delivery.on_settled(LinkDeliverySettleReason.TIMEOUT, None)
                continue
            if not delivery.sent:
                sent_and_settled = await self._outgoing_transfer(delivery)
                if sent_and_settled:
                    continue
            pending.append(delivery)
        self._pending_deliveries = pending

    async def send_transfer(self, message, *, send_async=False, **kwargs):
        self._check_if_closed()
        if self.state != LinkState.ATTACHED:
            raise AMQPLinkError(
                condition=ErrorCondition.ClientError,
                description="Link is not attached."
            )
        settled = self.send_settle_mode == SenderSettleMode.Settled
        if self.send_settle_mode == SenderSettleMode.Mixed:
            settled = kwargs.pop("settled", True)
        delivery = PendingDelivery(
            on_delivery_settled=kwargs.get("on_send_complete"),
            timeout=kwargs.get("timeout"),
            message=message,
            settled=settled,
            network_trace_params=self.network_trace_params
        )
        if self.current_link_credit == 0 or send_async:
            self._pending_deliveries.append(delivery)
        else:
            sent_and_settled = await self._outgoing_transfer(delivery)
            if not sent_and_settled:
                self._pending_deliveries.append(delivery)
        return delivery

    async def cancel_transfer(self, delivery):
        try:
            index = self._pending_deliveries.index(delivery)
        except ValueError:
            raise ValueError("Found no matching pending transfer.") from None
        delivery = self._pending_deliveries[index]
        if delivery.sent:
            raise MessageException(
                ErrorCondition.ClientError,
                message="Transfer cannot be cancelled. Message has already been sent and awaiting disposition.",
            )
        await delivery.on_settled(LinkDeliverySettleReason.CANCELLED, None)
        self._pending_deliveries.pop(index)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to show sending, receiving and parsing amqp annotated message(s) to Event Hubs.
"""

import os
import asyncio
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub.amqp import AmqpAnnotatedMessage, AmqpMessageBodyType


CONNECTION_STR = os.environ['EVENT_HUB_CONN_STR']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']

async def send_data_message(producer):
    data_body = [b'aa', b'bb', b'cc']
    application_properties = {"body_type": "data"}
    delivery_annotations = {"delivery_annotation_key": "value"}
    data_message = AmqpAnnotatedMessage(
        data_body=data_body,
        delivery_annotations=delivery_annotations,
        application_properties=application_properties
    )
    batch = await producer.create_batch()
    batch.add(data_message)
    await producer.send_batch(batch)
    print("Message of data body sent.")


async def send_sequence_message(producer):
    sequence_body = [b'message', 123.456, True]
    footer = {'footer_key': 'footer_value'}
    properties = {"subject": "sequence"}
    application_properties = {"body_type": "sequence"}
    sequence_message = AmqpAnnotatedMessage(
        sequence_body=sequence_body,
        footer=footer,
        properties=properties,
        application_properties=application_properties
    )
    await producer.send_batch([sequence_message])
    print("Message of sequence body sent.")


async def send_value_message(producer):
    value_body = {b"key": [-123, b'data', False]}
    header = {"priority": 10}
    annotations = {"annotation_key": "value"}
    application_properties = {"body_type": "value"}
    value_message = AmqpAnnotatedMessage(
        value_body=value_body,
        header=header,
        annotations=annotations,
        application_properties=application_properties
    )
    await producer.send_batch([value_message])
    print("Message of value body sent.")


async def on_event(partition_context, event):
    # Put your code here.
    # If the operation is i/o intensive, multi-thread will have better performance.
    print("Received event from partition: {}".format(partition_context.partition_id))
    raw_amqp_message = event.raw_amqp_message
    if raw_amqp_message.body_type == AmqpMessageBodyType.DATA:
        print("Message of data body received. Body is:")
        for data_section in raw_amqp_message.body:
            print(data_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.SEQUENCE:
        print("Message of sequence body received. Body is:")
        for sequence_section in raw_amqp_message.body:
            print(sequence_section)
    elif raw_amqp_message.body_type == AmqpMessageBodyType.VALUE:
        print("Message of value body received. Body is:")
        print(raw_amqp_message.body)


async def receive_and_parse_message(consumer):
    async with consumer:
        try:
            await consumer.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        except KeyboardInterrupt:
            print('Stopped receiving.')


async def main():
    # Send AmqpAnnotatedMessage
    producer = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME
    )
    async with producer:
        await send_data_message(producer)
        await send_sequence_message(producer)
        await send_value_message(producer)

    # Receive
    consumer = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
    )
    await receive_and_parse_message(consumer)

if __name__ == '__main__':
    asyncio.run(main())



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show sending and receiving events behind a proxy.
"""
import os
from azure.eventhub import EventData, EventHubConsumerClient, EventHubProducerClient

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']

HTTP_PROXY = {
    'proxy_hostname': '127.0.0.1',  # proxy hostname.
    'proxy_port': 3128,  # proxy port.
    'username': 'admin',  # username used for proxy authentication if needed.
    'password': '123456'  # password used for proxy authentication if needed.
}


def on_event(partition_context, event):
    # Put your code here.
    print("received event from partition: {}.".format(partition_context.partition_id))
    print(event)


consumer_client = EventHubConsumerClient.from_connection_string(
    conn_str=CONNECTION_STR, consumer_group='$Default', eventhub_name=EVENTHUB_NAME, http_proxy=HTTP_PROXY)
producer_client = EventHubProducerClient.from_connection_string(
    conn_str=CONNECTION_STR, eventhub_name=EVENTHUB_NAME, http_proxy=HTTP_PROXY)

with producer_client:
    event_data_batch = producer_client.create_batch(max_size_in_bytes=10000)
    while True:
        try:
            event_data_batch.add(EventData('Message inside EventBatchData'))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break
    producer_client.send_batch(event_data_batch)
    print('Finished sending.')

with consumer_client:
    consumer_client.receive(on_event=on_event)
    print('Finished receiving.')




#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub with checkpoint store doing checkpoint
by every fixed event count.
In the `receive` method of `EventHubConsumerClient`:
If no partition id is specified, the checkpoint_store are used for load-balance and checkpoint.
If partition id is specified, the checkpoint_store can only be used for checkpoint.
"""
from typing import Dict
import os
from azure.eventhub import EventHubConsumerClient
from azure.eventhub.extensions.checkpointstoreblob import BlobCheckpointStore


CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']
STORAGE_CONNECTION_STR = os.environ["AZURE_STORAGE_CONN_STR"]
BLOB_CONTAINER_NAME = "your-blob-container-name"  # Please make sure the blob container resource exists.

partition_recv_cnt_since_last_checkpoint: Dict[str, int] = {}
checkpoint_event_cnt = 20


def on_event(partition_context, event):
    # Put your code here.
    # Avoid time-consuming operations.
    p_id = partition_context.partition_id
    print("Received event from partition: {}".format(p_id))
    partition_recv_cnt_since_last_checkpoint[p_id] += 1
    if partition_recv_cnt_since_last_checkpoint[p_id] >= checkpoint_event_cnt:
        partition_context.update_checkpoint(event)
        partition_recv_cnt_since_last_checkpoint[p_id] = 0


if __name__ == '__main__':
    checkpoint_store = BlobCheckpointStore.from_connection_string(STORAGE_CONNECTION_STR, BLOB_CONTAINER_NAME)
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
        checkpoint_store=checkpoint_store,  # For load-balancing and checkpoint. Leave None for no load-balancing.
    )

    try:
        with consumer_client:
            """
            Without specified partition_id, the receive will try to receive events from all partitions and if provided
            with a checkpoint store, the client will load-balance partition assignment with other EventHubConsumerClient
            instances which also try to receive events from all partitions and use the same storage resource.
            """
            consumer_client.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
            # With specified partition_id, load-balance will be disabled, for example:
            # client.receive(on_event=on_event, partition_id='0')
    except KeyboardInterrupt:
        print('Stopped receiving.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import unicode_literals, annotations

import json
import warnings
import datetime
import logging
import uuid
from typing import (
    Union,
    Dict,
    Any,
    AnyStr,
    Iterable,
    Optional,
    List,
    TYPE_CHECKING,
    cast,
)
from typing_extensions import TypedDict

from ._utils import (
    utc_from_timestamp,
    transform_outbound_single_message,
    decode_with_recurse,
)
from ._tracing import trace_message
from ._constants import (
    PROP_SEQ_NUMBER,
    PROP_OFFSET,
    PROP_PARTITION_KEY,
    PROP_TIMESTAMP,
    PROP_ABSOLUTE_EXPIRY_TIME,
    PROP_CONTENT_ENCODING,
    PROP_CONTENT_TYPE,
    PROP_CORRELATION_ID,
    PROP_GROUP_ID,
    PROP_GROUP_SEQUENCE,
    PROP_MESSAGE_ID,
    PROP_REPLY_TO,
    PROP_REPLY_TO_GROUP_ID,
    PROP_SUBJECT,
    PROP_TO,
    PROP_USER_ID,
    PROP_CREATION_TIME,
)
from .amqp import (
    AmqpAnnotatedMessage,
    AmqpMessageBodyType,
    AmqpMessageHeader,
    AmqpMessageProperties,
)
from ._pyamqp._message_backcompat import LegacyMessage, LegacyBatchMessage
from ._pyamqp.message import Message as pyamqp_Message
from ._transport._pyamqp_transport import PyamqpTransport

if TYPE_CHECKING:
    try:
        from uamqp import (  # pylint: disable=unused-import
            Message,    # not importing as uamqp_Message, b/c type is exposed to user
            BatchMessage,
        )
    except ImportError:
        Message = None
        BatchMessage = None
    from ._transport._base import AmqpTransport

MessageContent = TypedDict("MessageContent", {"content": bytes, "content_type": str})
PrimitiveTypes = Optional[
    Union[
        int,
        float,
        bytes,
        bool,
        str,
        Dict,
        List,
        uuid.UUID,
    ]
]

_LOGGER = logging.getLogger(__name__)

# event_data.encoded_size < 255, batch encode overhead is 5, >=256, overhead is 8 each
_BATCH_MESSAGE_OVERHEAD_COST = [5, 8]

_SYS_PROP_KEYS_TO_MSG_PROPERTIES = (
    (PROP_MESSAGE_ID, "message_id"),
    (PROP_USER_ID, "user_id"),
    (PROP_TO, "to"),
    (PROP_SUBJECT, "subject"),
    (PROP_REPLY_TO, "reply_to"),
    (PROP_CORRELATION_ID, "correlation_id"),
    (PROP_CONTENT_TYPE, "content_type"),
    (PROP_CONTENT_ENCODING, "content_encoding"),
    (PROP_ABSOLUTE_EXPIRY_TIME, "absolute_expiry_time"),
    (PROP_CREATION_TIME, "creation_time"),
    (PROP_GROUP_ID, "group_id"),
    (PROP_GROUP_SEQUENCE, "group_sequence"),
    (PROP_REPLY_TO_GROUP_ID, "reply_to_group_id"),
)


class EventData(object):
    """The EventData class is a container for event content.

    :param body: The data to send in a single message. body can be type of str or bytes.
    :type body: str or bytes

    .. admonition:: Example:

        .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
            :start-after: [START create_event_data]
            :end-before: [END create_event_data]
            :language: python
            :dedent: 4
            :caption: Create instances of EventData

    """

    def __init__(
        self,
        body: Optional[Union[str, bytes, List[AnyStr]]] = None,
    ) -> None:
        self._last_enqueued_event_properties = {}  # type: Dict[str, Any]
        self._sys_properties = None  # type: Optional[Dict[bytes, Any]]
        if body is None:
            raise ValueError("EventData cannot be None.")

        # Internal usage only for transforming AmqpAnnotatedMessage to outgoing EventData
        self._raw_amqp_message = AmqpAnnotatedMessage(  # type: ignore
            data_body=body, annotations={}, application_properties={}
        )
        self._uamqp_message: Optional[Union[LegacyMessage, "Message"]] = None
        self._message: Union["Message", pyamqp_Message] = None  # type: ignore
        self._raw_amqp_message.header = AmqpMessageHeader()
        self._raw_amqp_message.properties = AmqpMessageProperties()
        self.message_id = None
        self.content_type = None
        self.correlation_id = None

    def __repr__(self) -> str:
        # pylint: disable=bare-except
        try:
            body_str = self.body_as_str()
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message body read error: %r", e)
            body_str = "<read-error>"
        event_repr = f"body='{body_str}'"
        try:
            event_repr += f", properties={self.properties}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message properties read error: %r", e)
            event_repr += ", properties=<read-error>"
        try:
            event_repr += f", offset={self.offset}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message offset read error: %r", e)
            event_repr += ", offset=<read-error>"
        try:
            event_repr += f", sequence_number={self.sequence_number}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message sequence number read error: %r", e)
            event_repr += ", sequence_number=<read-error>"
        try:
            event_repr += f", partition_key={self.partition_key!r}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message partition key read error: %r", e)
            event_repr += ", partition_key=<read-error>"
        try:
            event_repr += f", enqueued_time={self.enqueued_time!r}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message enqueued time read error: %r", e)
            event_repr += ", enqueued_time=<read-error>"
        return f"EventData({event_repr})"

    def __str__(self) -> str:
        try:
            body_str = self.body_as_str()
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message body read error: %r", e)
            body_str = "<read-error>"
        event_str = f"{{ body: '{body_str}'"
        try:
            event_str += f", properties: {self.properties}"
            if self.offset:
                event_str += f", offset: {self.offset}"
            if self.sequence_number:
                event_str += f", sequence_number: {self.sequence_number}"
            if self.partition_key:
                event_str += f", partition_key={self.partition_key!r}"
            if self.enqueued_time:
                event_str += f", enqueued_time={self.enqueued_time!r}"
        except Exception as e:  # pylint: disable=broad-except
            _LOGGER.debug("Message metadata read error: %r", e)
        event_str += " }"
        return event_str

    def __message_content__(self) -> MessageContent:
        if self.body_type != AmqpMessageBodyType.DATA:
            raise TypeError("`body_type` must be `AmqpMessageBodyType.DATA`.")
        content = bytearray()
        for c in self.body:  # type: ignore
            content += c  # type: ignore
        content_type = cast(str, self.content_type)
        return {"content": bytes(content), "content_type": content_type}

    @classmethod
    def from_message_content(  # pylint: disable=unused-argument
        cls, content: bytes, content_type: str, **kwargs: Any
    ) -> "EventData":
        """
        Creates an EventData object given content type and a content value to be set as body.

        :param bytes content: The content value to be set as the body of the message.
        :param str content_type: The content type to be set on the message.
        :return: An EventData object.
        :rtype: ~azure.eventhub.EventData
        """
        event_data = cls(content)
        event_data.content_type = content_type
        return event_data

    @classmethod
    def _from_message(
        cls,
        message: Union["Message", pyamqp_Message],
        raw_amqp_message: Optional[AmqpAnnotatedMessage] = None,
    ) -> EventData:
        # pylint:disable=protected-access
        """Internal use only.

        Creates an EventData object from a raw uamqp message and, if provided, AmqpAnnotatedMessage.

        :param ~uamqp.Message message: A received uamqp message.
        :param ~azure.eventhub.amqp.AmqpAnnotatedMessage message: An amqp annotated message.
        :param ~azure.eventhub.amqp.AmqpAnnotatedMessage raw_amqp_message: An amqp annotated message.
        :return: An EventData object.
        :rtype: ~azure.eventhub.EventData
        """
        event_data = cls(body="")
        # pylint: disable=protected-access
        event_data._message = message
        event_data._raw_amqp_message = (
            raw_amqp_message
            if raw_amqp_message
            else AmqpAnnotatedMessage(message=message)
        )
        return event_data

    def _decode_non_data_body_as_str(self, encoding: str = "UTF-8") -> str:
        # pylint: disable=protected-access
        body = self.raw_amqp_message.body
        if self.body_type == AmqpMessageBodyType.VALUE:
            if not body:
                return ""
            return str(decode_with_recurse(body, encoding))

        seq_list = [d for seq_section in body for d in seq_section]
        return str(decode_with_recurse(seq_list, encoding))

    @property
    def message(self) -> Union["Message", LegacyMessage]:
        """DEPRECATED: Get the underlying uamqp.Message or LegacyMessage.
         This is deprecated and will be removed in a later release.

        :rtype: uamqp.Message or LegacyMessage
        """
        warnings.warn(
            "The `message` property is deprecated and will be removed in future versions.",
            DeprecationWarning,
        )
        if not self._uamqp_message:
            self._uamqp_message = LegacyMessage(
                self._raw_amqp_message,
                to_outgoing_amqp_message=PyamqpTransport().to_outgoing_amqp_message,
            )
        return self._uamqp_message

    @message.setter
    def message(self, value: "Message") -> None:
        """DEPRECATED: Set the underlying Message.
        This is deprecated and will be removed in a later release.

        :param Message value: The message object.
        """
        warnings.warn(
            "The `message` property is deprecated and will be removed in future versions.",
            DeprecationWarning,
        )
        self._uamqp_message = value

    @property
    def raw_amqp_message(self) -> AmqpAnnotatedMessage:
        """Advanced usage only. The internal AMQP message payload that is sent or received.

        :rtype: ~azure.eventhub.amqp.AmqpAnnotatedMessage
        """
        return self._raw_amqp_message

    @property
    def sequence_number(self) -> Optional[int]:
        """The sequence number of the event.

        :rtype: int or None
        """
        return self._raw_amqp_message.annotations.get(PROP_SEQ_NUMBER, None)

    @property
    def offset(self) -> Optional[str]:
        """The offset of the event.

        :rtype: str or None
        """
        try:
            return self._raw_amqp_message.annotations[PROP_OFFSET].decode("UTF-8")
        except (KeyError, AttributeError):
            return None

    @property
    def enqueued_time(self) -> Optional[datetime.datetime]:
        """The enqueued timestamp of the event.

        :rtype: datetime.datetime or None
        """
        timestamp = self._raw_amqp_message.annotations.get(PROP_TIMESTAMP, None)
        if timestamp:
            return utc_from_timestamp(float(timestamp) / 1000)
        return None

    @property
    def partition_key(self) -> Optional[bytes]:
        """The partition key of the event.

        :rtype: bytes or None
        """
        return self._raw_amqp_message.annotations.get(PROP_PARTITION_KEY, None)

    @property
    def properties(self) -> Dict[Union[str, bytes], Any]:
        """Application-defined properties on the event.

        :rtype: dict[str, any] or dict[bytes, any]
        """
        return self._raw_amqp_message.application_properties

    @properties.setter
    def properties(self, value: Dict[Union[str, bytes], Any]):
        """Application-defined properties on the event.

        :param dict[str, any] or dict[bytes, any] value: The application properties for the EventData.
        """
        properties = None if value is None else dict(value)
        self._raw_amqp_message.application_properties = properties

    @property
    def system_properties(self) -> Dict[bytes, Any]:
        """Metadata set by the Event Hubs Service associated with the event.

        An EventData could have some or all of the following meta data depending on the source
        of the event data.

            - b"x-opt-sequence-number" (int)
            - b"x-opt-offset" (bytes)
            - b"x-opt-partition-key" (bytes)
            - b"x-opt-enqueued-time" (int)
            - b"message-id" (bytes)
            - b"user-id" (bytes)
            - b"to" (bytes)
            - b"subject" (bytes)
            - b"reply-to" (bytes)
            - b"correlation-id" (bytes)
            - b"content-type" (bytes)
            - b"content-encoding" (bytes)
            - b"absolute-expiry-time" (int)
            - b"creation-time" (int)
            - b"group-id" (bytes)
            - b"group-sequence" (bytes)
            - b"reply-to-group-id" (bytes)

        :rtype: dict[bytes, any]
        """

        if self._sys_properties is None:
            self._sys_properties = {}
            if self._raw_amqp_message.properties:
                for key, prop_name in _SYS_PROP_KEYS_TO_MSG_PROPERTIES:
                    value = getattr(self._raw_amqp_message.properties, prop_name, None)
                    if value:
                        self._sys_properties[key] = value
            self._sys_properties.update(self._raw_amqp_message.annotations)
        return self._sys_properties

    @property
    def body(self) -> PrimitiveTypes:
        """The body of the Message. The format may vary depending on the body type:
        For :class:`azure.eventhub.amqp.AmqpMessageBodyType.DATA<azure.eventhub.amqp.AmqpMessageBodyType.DATA>`,
        the body could be bytes or Iterable[bytes].
        For :class:`azure.eventhub.amqp.AmqpMessageBodyType.SEQUENCE<azure.eventhub.amqp.AmqpMessageBodyType.SEQUENCE>`,
        the body could be List or Iterable[List].
        For :class:`azure.eventhub.amqp.AmqpMessageBodyType.VALUE<azure.eventhub.amqp.AmqpMessageBodyType.VALUE>`,
        the body could be any type.

        :rtype: int or bool or float or bytes or str or dict or list or uuid.UUID
        """
        try:
            return self._raw_amqp_message.body
        except:
            raise ValueError("Event content empty.") from None

    @property
    def body_type(self) -> AmqpMessageBodyType:
        """The body type of the underlying AMQP message.

        :rtype: ~azure.eventhub.amqp.AmqpMessageBodyType
        """
        return self._raw_amqp_message.body_type

    def body_as_str(self, encoding: str = "UTF-8") -> str:
        """The content of the event as a string, if the data is of a compatible type.

        :param str encoding: The encoding to use for decoding event data.
         Default is 'UTF-8'
        :return: The content of the event as a string.
        :rtype: str
        """
        data = self.body
        try:
            if self.body_type != AmqpMessageBodyType.DATA:
                return self._decode_non_data_body_as_str(encoding=encoding)
            return "".join(b.decode(encoding) for b in cast(Iterable[bytes], data))
        except UnicodeDecodeError as e:
            raise TypeError(f"Message data is not compatible with string type: {e}") from e
        except TypeError:
            return str(data)
        except Exception:  # pylint: disable=broad-except
            pass
        try:
            return cast(bytes, data).decode(encoding)
        except Exception as e:
            raise TypeError(f"Message data is not compatible with string type: {e}") from e

    def body_as_json(self, encoding: str = "UTF-8") -> Dict[str, Any]:
        """The content of the event loaded as a JSON object, if the data is compatible.

        :param str encoding: The encoding to use for decoding event data.
         Default is 'UTF-8'
        :return: A JSON object.
        :rtype: dict[str, any]
        """
        data_str = self.body_as_str(encoding=encoding)
        try:
            return json.loads(data_str)
        except Exception as e:
            raise TypeError(f"Event data is not compatible with JSON type: {e}") from e

    @property
    def content_type(self) -> Optional[str]:
        """The content type descriptor.
        Optionally describes the payload of the message, with a descriptor following the format of RFC2045, Section 5,
        for example "application/json".
        :rtype: str or None
        """
        if not self._raw_amqp_message.properties:
            return None
        try:
            return self._raw_amqp_message.properties.content_type.decode("UTF-8")
        except (AttributeError, UnicodeDecodeError):
            return self._raw_amqp_message.properties.content_type

    @content_type.setter
    def content_type(self, value: str) -> None:
        if not self._raw_amqp_message.properties:
            self._raw_amqp_message.properties = AmqpMessageProperties()
        self._raw_amqp_message.properties.content_type = value

    @property
    def correlation_id(self) -> Optional[str]:
        """The correlation identifier.
        Allows an application to specify a context for the message for the purposes of correlation, for example
        reflecting the MessageId of a message that is being replied to.
        :rtype: str or None
        """
        if not self._raw_amqp_message.properties:
            return None
        try:
            return self._raw_amqp_message.properties.correlation_id.decode("UTF-8")
        except (AttributeError, UnicodeDecodeError):
            return self._raw_amqp_message.properties.correlation_id

    @correlation_id.setter
    def correlation_id(self, value: str) -> None:
        if not self._raw_amqp_message.properties:
            self._raw_amqp_message.properties = AmqpMessageProperties()
        self._raw_amqp_message.properties.correlation_id = value

    @property
    def message_id(self) -> Optional[str]:
        """The id to identify the message.
        The message identifier is an application-defined value that uniquely identifies the message and its payload.
        The identifier is a free-form string and can reflect a GUID or an identifier derived from the
        application context.  If enabled, the duplicate detection feature identifies and removes second and
        further submissions of messages with the same message id.
        :rtype: str or None
        """
        if not self._raw_amqp_message.properties:
            return None
        try:
            return self._raw_amqp_message.properties.message_id.decode("UTF-8")
        except (AttributeError, UnicodeDecodeError):
            return self._raw_amqp_message.properties.message_id

    @message_id.setter
    def message_id(self, value: str) -> None:
        if not self._raw_amqp_message.properties:
            self._raw_amqp_message.properties = AmqpMessageProperties()
        self._raw_amqp_message.properties.message_id = value


class EventDataBatch(object):
    """A batch of events.

    Sending events in a batch is more performant than sending individual events.
    EventDataBatch helps you create the maximum allowed size batch of `EventData` to improve sending performance.

    Use the `add` method to add events until the maximum batch size limit in bytes has been reached -
    at which point a `ValueError` will be raised.
    Use the `send_batch` method of :class:`EventHubProducerClient<azure.eventhub.EventHubProducerClient>`
    or the async :class:`EventHubProducerClient<azure.eventhub.aio.EventHubProducerClient>`
    for sending.

    **Please use the create_batch method of EventHubProducerClient
    to create an EventDataBatch object instead of instantiating an EventDataBatch object directly.**

    **WARNING: Updating the value of the instance variable max_size_in_bytes on an instantiated EventDataBatch object
    is HIGHLY DISCOURAGED. The updated max_size_in_bytes value may conflict with the maximum size of events allowed
    by the Event Hubs service and result in a sending failure.**

    :param int max_size_in_bytes: The maximum size of bytes data that an EventDataBatch object can hold.
    :param str partition_id: The specific partition ID to send to.
    :param str partition_key: With the given partition_key, event data will be sent to a particular partition of the
     Event Hub decided by the service.
    """

    def __init__(
        self,
        max_size_in_bytes: Optional[int] = None,
        partition_id: Optional[str] = None,
        partition_key: Optional[Union[str, bytes]] = None,
        **kwargs,
    ) -> None:
        self._amqp_transport = kwargs.pop("amqp_transport", PyamqpTransport)
        self._tracing_attributes: Dict[str, Union[str, int]] = kwargs.pop("tracing_attributes", {})

        if partition_key and not isinstance(partition_key, (str, bytes)):
            _LOGGER.info(
                "WARNING: Setting partition_key of non-string value on the events to be sent is discouraged "
                "as the partition_key will be ignored by the Event Hub service and events will be assigned "
                "to all partitions using round-robin. Furthermore, there are SDKs for consuming events which expect "
                "partition_key to only be string type, they might fail to parse the non-string value."
            )

        self._partition_id = partition_id
        self._partition_key = partition_key

        self._message = self._amqp_transport.build_batch_message(data=[])
        self._message = self._amqp_transport.set_message_partition_key(
            self._message, self._partition_key
        )
        self._size = self._amqp_transport.get_batch_message_encoded_size(self._message)
        self.max_size_in_bytes = (
            max_size_in_bytes or self._amqp_transport.MAX_MESSAGE_LENGTH_BYTES
        )

        self._count = 0
        self._internal_events: List[Union[EventData, AmqpAnnotatedMessage]] = []
        self._uamqp_message: Optional[Union[BatchMessage, LegacyBatchMessage]] = None
        if self._amqp_transport.KIND == "uamqp":
            self._uamqp_message = self._message

    def __repr__(self) -> str:
        batch_repr = (
            f"max_size_in_bytes={self.max_size_in_bytes}, partition_id={self._partition_id}, "
            f"partition_key={self._partition_key!r}, event_count={self._count}"
        )
        return f"EventDataBatch({batch_repr})"

    def __len__(self) -> int:
        return self._count

    @classmethod
    def _from_batch(
        cls,
        batch_data: Iterable[Union[AmqpAnnotatedMessage, EventData]],
        amqp_transport: AmqpTransport,
        partition_key: Optional[AnyStr] = None,
        *,
        max_size_in_bytes: Optional[int] = None,
        partition_id: Optional[str] = None,
    ) -> EventDataBatch:
        outgoing_batch_data = [
            transform_outbound_single_message(
                m, EventData, amqp_transport.to_outgoing_amqp_message
            )
            for m in batch_data
        ]
        batch_data_instance = cls(
            partition_key=partition_key,
            amqp_transport=amqp_transport,
            max_size_in_bytes=max_size_in_bytes,
            partition_id=partition_id,
        )

        for event_data in outgoing_batch_data:
            batch_data_instance.add(event_data)
        return batch_data_instance

    def _load_events(self, events):
        for event_data in events:
            try:
                self.add(event_data)
            except ValueError:
                raise ValueError(
                    "The combined size of EventData or AmqpAnnotatedMessage collection exceeds "
                    "the Event Hub frame size limit. Please send a smaller collection of EventData "
                    "or use EventDataBatch, which is guaranteed to be under the frame size limit"
                ) from None

    @property
    def message(self) -> Union["BatchMessage", LegacyBatchMessage]:
        """DEPRECATED: Get the underlying uamqp.BatchMessage or LegacyBatchMessage.
         This is deprecated and will be removed in a later release.

        :rtype: uamqp.BatchMessage or LegacyBatchMessage
        """
        warnings.warn(
            "The `message` property is deprecated and will be removed in future versions.",
            DeprecationWarning,
        )
        if not self._uamqp_message:
            message = AmqpAnnotatedMessage(message=pyamqp_Message(*self._message))
            self._uamqp_message = LegacyBatchMessage(
                message,
                to_outgoing_amqp_message=PyamqpTransport().to_outgoing_amqp_message,
            )
        return self._uamqp_message

    @message.setter
    def message(self, value: "BatchMessage") -> None:
        """DEPRECATED: Set the underlying BatchMessage.
        This is deprecated and will be removed in a later release.

        :param BatchMessage value: The BatchMessage to set as the underlying amqp message for the EventDataBatch.
        """
        warnings.warn(
            "The `message` property is deprecated and will be removed in future versions.",
            DeprecationWarning,
        )
        self._uamqp_message = value

    @property
    def size_in_bytes(self) -> int:
        """The combined size of the events in the batch, in bytes.

        :rtype: int
        """
        return self._size

    def add(self, event_data: Union[EventData, AmqpAnnotatedMessage]) -> None:
        """Try to add an EventData to the batch.

        The total size of an added event is the sum of its body, properties, etc.
        If this added size results in the batch exceeding the maximum batch size, a `ValueError` will
        be raised.

        :param event_data: The EventData to add to the batch.
        :type event_data: Union[~azure.eventhub.EventData, ~azure.eventhub.amqp.AmqpAnnotatedMessage]
        :rtype: None
        :raise: :class:`ValueError`, when exceeding the size limit.
        """

        outgoing_event_data = transform_outbound_single_message(
            event_data, EventData, self._amqp_transport.to_outgoing_amqp_message
        )

        if self._partition_key:
            if (
                outgoing_event_data.partition_key
                and outgoing_event_data.partition_key != self._partition_key
            ):
                raise ValueError(
                    "The partition key of event_data does not match the partition key of this batch."
                )
            if not outgoing_event_data.partition_key:
                outgoing_event_data._message = self._amqp_transport.set_message_partition_key(  # pylint: disable=protected-access
                    outgoing_event_data._message,  # pylint: disable=protected-access
                    self._partition_key,
                )

        outgoing_event_data._message = trace_message(   # pylint: disable=protected-access
            outgoing_event_data._message,   # pylint: disable=protected-access
            amqp_transport=self._amqp_transport,
            additional_attributes=self._tracing_attributes
        )
        event_data_size = self._amqp_transport.get_message_encoded_size(
            outgoing_event_data._message  # pylint: disable=protected-access
        )
        # For a BatchMessage, if the encoded_message_size of event_data is < 256, then the overhead cost to encode that
        # message into the BatchMessage would be 5 bytes, if >= 256, it would be 8 bytes.
        size_after_add = (
            self._size
            + event_data_size
            + _BATCH_MESSAGE_OVERHEAD_COST[0 if (event_data_size < 256) else 1]
        )

        if size_after_add > self.max_size_in_bytes:
            raise ValueError(
                f"EventDataBatch has reached its size limit: {self.max_size_in_bytes}"
            )

        self._amqp_transport.add_batch(self, outgoing_event_data, event_data)
        self._size = size_after_add
        self._count += 1



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._mixin import DictMixin
from ._client_base import _parse_conn_str


class EventHubConnectionStringProperties(DictMixin):
    """
    Properties of a connection string.
    """

    def __init__(self, **kwargs):
        self._fully_qualified_namespace = kwargs.pop("fully_qualified_namespace", None)
        self._endpoint = kwargs.pop("endpoint", None)
        self._eventhub_name = kwargs.pop("eventhub_name", None)
        self._shared_access_signature = kwargs.pop("shared_access_signature", None)
        self._shared_access_key_name = kwargs.pop("shared_access_key_name", None)
        self._shared_access_key = kwargs.pop("shared_access_key", None)

    @property
    def fully_qualified_namespace(self):
        """The fully qualified host name for the Event Hubs namespace.
        The namespace format is: `<yournamespace>.servicebus.windows.net`.

        :rtype: str
        """
        return self._fully_qualified_namespace

    @property
    def endpoint(self):
        """The endpoint for the Event Hubs resource. In the format sb://<FQDN>/

        :rtype: str
        """
        return self._endpoint

    @property
    def eventhub_name(self):
        """Optional. The name of the Event Hub, represented by `EntityPath` in the connection string.

        :rtype: str
        """
        return self._eventhub_name

    @property
    def shared_access_signature(self):
        """
        This can be provided instead of the shared_access_key_name and the shared_access_key.

        :rtype: str
        """
        return self._shared_access_signature

    @property
    def shared_access_key_name(self):
        """
        The name of the shared_access_key. This must be used along with the shared_access_key.

        :rtype: str
        """
        return self._shared_access_key_name

    @property
    def shared_access_key(self):
        """
        The shared_access_key can be used along with the shared_access_key_name as a credential.

        :rtype: str
        """
        return self._shared_access_key


def parse_connection_string(conn_str):
    # type(str) -> EventHubConnectionStringProperties
    """Parse the connection string into a properties bag containing its component parts.

    :param conn_str: The connection string that has to be parsed.
    :type conn_str: str
    :return: A properties bag containing the parsed connection string.
    :rtype: ~azure.eventhub.EventHubConnectionStringProperties
    """
    fully_qualified_namespace, policy, key, entity, signature = _parse_conn_str(
        conn_str, check_case=True
    )[:-1]
    endpoint = "sb://" + fully_qualified_namespace + "/"
    props = {
        "fully_qualified_namespace": fully_qualified_namespace,
        "endpoint": endpoint,
        "eventhub_name": entity,
        "shared_access_signature": signature,
        "shared_access_key_name": policy,
        "shared_access_key": key,
    }
    return EventHubConnectionStringProperties(**props)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import time
from collections import defaultdict
from threading import Thread
from uuid import uuid4
from concurrent.futures import ThreadPoolExecutor

import pytest

from azure.eventhub import EventData
from azure.eventhub import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub._buffered_producer import PartitionResolver
from azure.eventhub.amqp import (
    AmqpAnnotatedMessage,
)
from azure.eventhub.exceptions import EventDataSendError, OperationTimeoutError, EventHubError


def random_pkey_generation(partitions):
    pr = PartitionResolver(partitions)
    total = len(partitions)
    dic = {}

    while total:
        key = str(uuid4())
        pid = pr.get_partition_id_by_partition_key(key)
        if pid in dic:
            continue
        else:
            dic[pid] = key
            total -= 1

    return dic


@pytest.mark.liveTest()
def test_producer_client_constructor(connection_str, uamqp_transport):
    def on_success(events, pid):
        pass

    def on_error(events, error, pid):
        pass

    with pytest.raises(TypeError):
        EventHubProducerClient.from_connection_string(connection_str, buffered_mode=True, uamqp_transport=uamqp_transport)
    with pytest.raises(TypeError):
        EventHubProducerClient.from_connection_string(connection_str, buffered_mode=True, on_success=on_success, uamqp_transport=uamqp_transport)
    with pytest.raises(TypeError):
        EventHubProducerClient.from_connection_string(connection_str, buffered_mode=True, on_error=on_error, uamqp_transport=uamqp_transport)
    with pytest.raises(ValueError):
        EventHubProducerClient.from_connection_string(
            connection_str,
            buffered_mode=True,
            on_success=on_success,
            on_error=on_error,
            max_wait_time=0,
            uamqp_transport=uamqp_transport
        )
    with pytest.raises(ValueError):
        EventHubProducerClient.from_connection_string(
            connection_str,
            buffered_mode=True,
            on_success=on_success,
            on_error=on_error,
            max_buffer_length=0,
            uamqp_transport=uamqp_transport
        )

    def on_success_missing_params(events):
        on_success_missing_params.events = events
    
    def on_error_missing_params(events, pid):
        on_error_missing_params.events = events

    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        buffer_concurrency=2,
        on_success=on_success_missing_params,
        on_error=on_error_missing_params,
        uamqp_transport=uamqp_transport,
    )
    
    on_success_missing_params.events = None
    on_error_missing_params.events = None

    # successfully send, but don't enter invalid callback
    with producer:
        producer.send_event(EventData('Single data'))
    
    assert not on_success_missing_params.events
    assert not on_error_missing_params.events

@pytest.mark.liveTest
@pytest.mark.parametrize(
    "flush_after_sending, close_after_sending",
    [
        (False, False),
        (True, False),
        (False, True)
    ]
)
@pytest.mark.liveTest
def test_basic_send_single_events_round_robin(connection_str, flush_after_sending, close_after_sending, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()

    time.sleep(10)
    sent_events = defaultdict(list)

    def on_success(events, pid):
        if len(events) > 1:
            on_success.batching = True
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    on_success.batching = False  # ensure batching happened

    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        partitions = producer.get_partition_ids()
        partitions_cnt = len(partitions)
        # perform single sending round-robin
        total_single_event_cnt = 100
        eventdata_set, amqpannoated_set = set(), set()
        for i in range(total_single_event_cnt // 2):
            event = EventData("test:{}".format(i))
            event.properties = {"event_idx": i}
            producer.send_event(event)
            eventdata_set.add(i)
        for i in range(total_single_event_cnt // 2, total_single_event_cnt):
            event = AmqpAnnotatedMessage(data_body="test:{}".format(i))
            event.application_properties = {"event_idx": i}
            amqpannoated_set.add(i)
            producer.send_event(event)

        for pid in partitions:
            assert producer.get_buffered_event_count(pid) > 0
        assert producer.total_buffered_event_count > 0

        if not flush_after_sending and not close_after_sending:
            # ensure it's buffered sending
            for pid in partitions:
                assert len(sent_events[pid]) < total_single_event_cnt // partitions_cnt
            assert sum([len(sent_events[pid]) for pid in partitions]) < total_single_event_cnt
        else:
            if flush_after_sending:
                producer.flush()
            if close_after_sending:
                producer.close()
            # ensure all events are sent after calling flush
            assert sum([len(sent_events[pid]) for pid in partitions]) == total_single_event_cnt

        # give some time for producer to complete sending and consumer to complete receiving
        time.sleep(10)
        assert len(sent_events) == len(received_events) == partitions_cnt

        for pid in partitions:
            assert producer.get_buffered_event_count(pid) == 0
        assert producer.total_buffered_event_count == 0
        assert not on_error.err

        # ensure all events are received in the correct partition
        for pid in partitions:
            assert len(sent_events[pid]) >= total_single_event_cnt // partitions_cnt
            assert len(sent_events[pid]) == len(received_events[pid])
            for i in range(len(sent_events[pid])):
                event = sent_events[pid][i]
                try:  # amqp annotated message
                    event_idx = event.application_properties["event_idx"]
                    amqpannoated_set.remove(event_idx)
                except AttributeError:  # event data
                    event_idx = event.properties["event_idx"]
                    eventdata_set.remove(event_idx)
                assert received_events[pid][i].properties[b"event_idx"] == event_idx
                assert partitions[event_idx % partitions_cnt] == pid

        assert on_success.batching
        assert not eventdata_set
        assert not amqpannoated_set

    consumer.close()
    receive_thread.join()


@pytest.mark.liveTest
@pytest.mark.parametrize(
    "flush_after_sending, close_after_sending",
    [
        (False, False),
        (True, False),
        (False, True)
    ]
)
def test_basic_send_batch_events_round_robin(connection_str, flush_after_sending, close_after_sending, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()

    time.sleep(10)
    sent_events = defaultdict(list)

    def on_success(events, pid):
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        partitions = producer.get_partition_ids()
        partitions_cnt = len(partitions)
        # perform batch sending round-robin
        total_events_cnt = 100
        batch_cnt = partitions_cnt * 2 - 1
        each_partition_cnt = total_events_cnt // batch_cnt
        remain_events = total_events_cnt % batch_cnt
        batches = []
        event_idx = 0
        eventdata_set, amqpannoated_set = set(), set()
        for i in range(batch_cnt):
            batch = producer.create_batch()
            for j in range(each_partition_cnt // 2):
                event = EventData("test{}:{}".format(i, event_idx))
                event.properties = {'batch_idx': i, 'event_idx': event_idx}
                batch.add(event)
                eventdata_set.add(event_idx)
                event_idx += 1
            for j in range(each_partition_cnt // 2, each_partition_cnt):
                event = AmqpAnnotatedMessage(data_body="test{}:{}".format(i, event_idx))
                event.application_properties = {'batch_idx': i, 'event_idx': event_idx}
                batch.add(event)
                amqpannoated_set.add(event_idx)
                event_idx += 1
            batches.append(batch)

        # put remain_events in the last batch
        last_batch = producer.create_batch()
        for i in range(remain_events):
            event = EventData("test:{}:{}".format(len(batches), event_idx))
            event.properties = {'batch_idx': len(batches), 'event_idx': event_idx}
            last_batch.add(event)
            eventdata_set.add(event_idx)
            event_idx += 1
        batches.append(last_batch)

        for batch in batches:
            producer.send_batch(batch)

        if not flush_after_sending and not close_after_sending:
            # ensure it's buffered sending
            for pid in partitions:
                assert len(sent_events[pid]) < each_partition_cnt
            assert sum([len(sent_events[pid]) for pid in partitions]) < total_events_cnt
            # give some time for producer to complete sending and consumer to complete receiving
        else:
            if flush_after_sending:
                producer.flush()
            if close_after_sending:
                producer.close()
            # ensure all events are sent
            assert sum([len(sent_events[pid]) for pid in partitions]) == total_events_cnt

        time.sleep(20)
        assert len(sent_events) == len(received_events) == partitions_cnt

        # ensure all events are received in the correct partition
        for pid in partitions:
            assert len(sent_events[pid]) > 0
            assert len(sent_events[pid]) == len(received_events[pid])
            for i in range(len(sent_events[pid])):
                event = sent_events[pid][i]
                try:  # amqp annotated message
                    event_idx = event.application_properties["event_idx"]
                    amqpannoated_set.remove(event_idx)
                except AttributeError:  # event data
                    event_idx = event.properties["event_idx"]
                    eventdata_set.remove(event_idx)
                assert received_events[pid][i].properties[b"event_idx"] == event_idx

        assert not amqpannoated_set
        assert not eventdata_set
        assert not on_error.err

    consumer.close()
    receive_thread.join()


@pytest.mark.liveTest
def test_send_with_hybrid_partition_assignment(connection_str, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()

    time.sleep(5)
    sent_events = defaultdict(list)

    def on_success(events, pid):
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        partitions = producer.get_partition_ids()
        partitions_cnt = len(partitions)
        pid_to_pkey = random_pkey_generation(partitions)
        expected_event_idx_to_partition = {}
        event_idx = 0
        # 1. send by partition_key, each partition 2 events, two single + one batch containing two
        for pid in partitions:
            pkey = pid_to_pkey[pid]
            producer.send_event(EventData('{}'.format(event_idx)), partition_key=pkey)
            batch = producer.create_batch(partition_key=pkey)
            batch.add(EventData('{}'.format(event_idx + 1)))
            producer.send_batch(batch)
            for i in range(2):
                expected_event_idx_to_partition[event_idx + i] = pid
            event_idx += 2

        # 2. send by partition_id, each partition 2 events, two single + one batch containing two
        for pid in partitions:
            producer.send_event(EventData('{}'.format(event_idx)), partition_id=pid)
            batch = producer.create_batch(partition_id=pid)
            batch.add(EventData('{}'.format(event_idx + 1)))
            producer.send_batch(batch)
            for i in range(2):
                expected_event_idx_to_partition[event_idx + i] = pid
            event_idx += 2

        # 3. send without partition, each partition 2 events, two single + one batch containing two
        for _ in partitions:
            producer.send_event(EventData('{}'.format(event_idx)))
            batch = producer.create_batch()
            batch.add(EventData('{}'.format(event_idx + 1)))
            producer.send_batch(batch)
            event_idx += 2

        producer.flush()
        assert len(sent_events) == partitions_cnt

        time.sleep(10)

        visited = set()
        for pid in partitions:
            assert len(sent_events[pid]) == 2 * 3

            for sent_event in sent_events[pid]:
                if int(sent_event.body_as_str()) in expected_event_idx_to_partition:
                    assert expected_event_idx_to_partition[int(sent_event.body_as_str())] == pid

            for recv_event in received_events[pid]:
                if int(sent_event.body_as_str()) in expected_event_idx_to_partition:
                    assert expected_event_idx_to_partition[int(sent_event.body_as_str())] == pid

                assert recv_event.body_as_str() not in visited
                visited.add(recv_event.body_as_str())

        assert len(visited) == 2 * 3 * len(partitions)

    assert not on_error.err
    consumer.close()
    receive_thread.join()


def test_send_with_timing_configuration(connection_str, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()

    time.sleep(5)
    sent_events = defaultdict(list)

    def on_success(events, pid):
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None

    # test max_wait_time
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        max_wait_time=10,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        partitions = producer.get_partition_ids()
        producer.send_batch([EventData('data')])
        time.sleep(5)
        assert not sent_events
        time.sleep(10)
        assert sum([len(sent_events[pid]) for pid in partitions]) == 1

    assert not on_error.err

    # test max_buffer_length per partition
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        max_wait_time=1000,
        max_buffer_length=10,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    sent_events.clear()
    received_events.clear()
    with producer:
        partitions = producer.get_partition_ids()
        for i in range(7):
            producer.send_event(EventData('data'), partition_id="0")
        assert not sent_events
        batch = producer.create_batch(partition_id="0")
        for i in range(9):
            batch.add(EventData('9'))
        producer.send_batch(batch)  # will flush 7 events and put the batch in buffer
        assert sum([len(sent_events[pid]) for pid in partitions]) == 7
        for i in range(5):
            producer.send_event(EventData('data'), partition_id="0")  # will flush batch (9 events) + 1 event, leaving 4 in buffer
        assert sum([len(sent_events[pid]) for pid in partitions]) == 17
        producer.flush()
        assert sum([len(sent_events[pid]) for pid in partitions]) == 21

    time.sleep(5)
    assert sum([len(received_events[pid]) for pid in partitions]) == 21
    assert not on_error.err
    consumer.close()
    receive_thread.join()


@pytest.mark.liveTest
def test_long_sleep(connection_str, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()

    time.sleep(5)
    sent_events = defaultdict(list)

    def on_success(events, pid):
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport
    )

    with producer:
        producer.send_event(EventData("test"), partition_id="0")
        time.sleep(220)
        producer.send_event(EventData("test"), partition_id="0")
        time.sleep(5)

    assert not on_error.err
    assert len(sent_events["0"]) == 2
    assert len(received_events["0"]) == 2

    consumer.close()
    receive_thread.join()

@pytest.mark.skip('not testing correctly + flaky, fix during MQ')
@pytest.mark.liveTest
def test_long_wait_small_buffer(connection_str, uamqp_transport):
    received_events = defaultdict(list)

    def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient.from_connection_string(connection_str, consumer_group="$default", uamqp_transport=uamqp_transport)
    receive_thread = Thread(target=consumer.receive, args=(on_event,))
    receive_thread.daemon = True
    receive_thread.start()
    time.sleep(10)

    sent_events = defaultdict(list)

    def on_success(events, pid):
        sent_events[pid].extend(events)

    def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    producer = EventHubProducerClient.from_connection_string(
        connection_str,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        auth_timeout=3, 
        retry_total=3, 
        retry_mode='fixed',
        retry_backoff_factor=0.01,
        max_wait_time=10,
        max_buffer_length=100,
        uamqp_transport=uamqp_transport
    )

    with producer:
        for i in range(100):
            producer.send_event(EventData("test"))

    time.sleep(60)

    assert not on_error.err
    assert sum([len(sent_events[key]) for key in sent_events]) == 100
    assert sum([len(received_events[key]) for key in received_events]) == 100

    consumer.close()
    receive_thread.join()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub for a period of time.
"""
import os
import threading
import time
from azure.eventhub import EventHubConsumerClient

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']
RECEIVE_DURATION = 15


def on_event(partition_context, event):
    # Put your code here.
    print("Received event from partition: {}.".format(partition_context.partition_id))


def on_partition_initialize(partition_context):
    # Put your code here.
    print("Partition: {} has been initialized.".format(partition_context.partition_id))


def on_partition_close(partition_context, reason):
    # Put your code here.
    print("Partition: {} has been closed, reason for closing: {}.".format(
        partition_context.partition_id,
        reason
    ))


def on_error(partition_context, error):
    # Put your code here. partition_context can be None in the on_error callback.
    if partition_context:
        print("An exception: {} occurred during receiving from Partition: {}.".format(
            partition_context.partition_id,
            error
        ))
    else:
        print("An exception: {} occurred during the load balance process.".format(error))


if __name__ == '__main__':
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
    )

    print('Consumer will keep receiving for {} seconds, start time is {}.'.format(RECEIVE_DURATION, time.time()))

    try:
        thread = threading.Thread(
            target=consumer_client.receive,
            kwargs={
                "on_event": on_event,
                "on_partition_initialize": on_partition_initialize,
                "on_partition_close": on_partition_close,
                "on_error": on_error,
                "starting_position": "-1",  # "-1" is from the beginning of the partition.
            }
        )
        thread.daemon = True
        thread.start()
        time.sleep(RECEIVE_DURATION)
        consumer_client.close()
        thread.join()
    except KeyboardInterrupt:
        print('Stop receiving.')

    print('Consumer has stopped receiving, end time is {}.'.format(time.time()))



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events in buffered mode to an Event Hub asynchronously.
"""

import time
import asyncio
import os

from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub import EventData

CONNECTION_STR = os.environ['EVENT_HUB_CONN_STR']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']


async def on_success(events, pid):
    # sending succeeded
    print(events, pid)


async def on_error(events, pid, error):
    # sending failed
    print(events, pid, error)


async def run():

    producer = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME,
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error
    )

    # exiting the context manager will automatically call flush
    async with producer:
        # single events will be batched automatically
        for i in range(10):
            # the method returning indicates the event has been enqueued to the buffer
            await producer.send_event(EventData('Single data {}'.format(i)))

        batch = await producer.create_batch()
        for i in range(10):
            batch.add(EventData('Single data in batch {}'.format(i)))
        # alternatively, you can enqueue an EventDataBatch object to the buffer
        await producer.send_batch(batch)

        # calling flush sends out the events in the buffer immediately
        await producer.flush()

start_time = time.time()
asyncio.run(run())
print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
import asyncio
import logging
import queue
import time
from asyncio import Lock
from typing import Optional, Callable, Awaitable, TYPE_CHECKING

from .._producer_async import EventHubProducer
from ..._common import EventDataBatch
from ...exceptions import OperationTimeoutError

if TYPE_CHECKING:
    from .._transport._base_async import AmqpTransportAsync
    from ..._producer_client import SendEventTypes

_LOGGER = logging.getLogger(__name__)


class BufferedProducer:
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        producer: EventHubProducer,
        partition_id: str,
        on_success: Callable[["SendEventTypes", Optional[str]], Awaitable[None]],
        on_error: Callable[
            ["SendEventTypes", Optional[str], Exception], Awaitable[None]
        ],
        max_message_size_on_link: int,
        *,
        amqp_transport: AmqpTransportAsync,
        max_buffer_length: int,
        max_wait_time: float = 1
    ):
        self._buffered_queue: queue.Queue = queue.Queue()
        self._max_buffer_len = max_buffer_length
        self._cur_buffered_len = 0
        self._producer: EventHubProducer = producer
        self._lock = Lock()
        self._max_wait_time = max_wait_time
        self._on_success = self.failsafe_callback(on_success)
        self._on_error = self.failsafe_callback(on_error)
        self._last_send_time = None
        self._running = False
        self._cur_batch: Optional[EventDataBatch] = None
        self._max_message_size_on_link = max_message_size_on_link
        self._check_max_wait_time_future = None
        self.partition_id = partition_id
        self._amqp_transport = amqp_transport

    async def start(self):
        async with self._lock:
            self._cur_batch = EventDataBatch(self._max_message_size_on_link, amqp_transport=self._amqp_transport)
            self._running = True
            if self._max_wait_time:
                self._last_send_time = time.time()
                self._check_max_wait_time_future = asyncio.ensure_future(
                    self.check_max_wait_time_worker()
                )

    async def stop(self, flush=True, timeout_time=None, raise_error=False):
        self._running = False
        if flush:
            async with self._lock:
                await self._flush(timeout_time=timeout_time, raise_error=raise_error)
        else:
            if self._cur_buffered_len:
                _LOGGER.warning(
                    "Shutting down Partition %r."
                    " There are still %r events in the buffer which will be lost",
                    self.partition_id,
                    self._cur_buffered_len,
                )
        if self._check_max_wait_time_future:
            try:
                await self._check_max_wait_time_future
            except Exception as exc:  # pylint: disable=broad-except
                _LOGGER.warning(
                    "Partition %r stopped with error %r", self.partition_id, exc
                )
        await self._producer.close()

    async def put_events(self, events, timeout_time=None):
        # Put single event or EventDataBatch into the queue.
        # This method would raise OperationTimeout if the queue does not have enough space for the input and
        # flush cannot finish in timeout.
        try:
            new_events_len = len(events)
        except TypeError:
            new_events_len = 1
        if self._max_buffer_len - self._cur_buffered_len < new_events_len:
            _LOGGER.info(
                "The buffer for partition %r is full. Attempting to flush before adding %r events.",
                self.partition_id,
                new_events_len,
            )
            # flush the buffer
            await self.flush(timeout_time=timeout_time)
        if timeout_time and time.time() > timeout_time:
            raise OperationTimeoutError(
                "Failed to enqueue events into buffer due to timeout."
            )
        try:
            # add single event into current batch
            self._cur_batch.add(events)
        except AttributeError:  # if the input events is a EventDataBatch, put the whole into the buffer
            # if there are events in cur_batch, enqueue cur_batch to the buffer
            async with self._lock:
                if self._cur_batch:
                    self._buffered_queue.put(self._cur_batch)
                self._buffered_queue.put(events)
            # create a new batch for incoming events
            self._cur_batch = EventDataBatch(self._max_message_size_on_link, amqp_transport=self._amqp_transport)
        except ValueError:
            # add single event exceeds the cur batch size, create new batch
            async with self._lock:
                self._buffered_queue.put(self._cur_batch)
            self._cur_batch = EventDataBatch(self._max_message_size_on_link, amqp_transport=self._amqp_transport)
            self._cur_batch.add(events)
        async with self._lock:
            self._cur_buffered_len += new_events_len

    def failsafe_callback(self, callback):
        async def wrapper_callback(*args, **kwargs):
            try:
                await callback(*args, **kwargs)
            except Exception as exc:  # pylint: disable=broad-except
                _LOGGER.warning(
                    "On partition %r, callback %r encountered exception %r",
                    callback.__name__,
                    exc,
                    self.partition_id,
                )

        return wrapper_callback

    async def flush(self, timeout_time=None, raise_error=True):
        async with self._lock:
            await self._flush(timeout_time, raise_error)

    async def _flush(self, timeout_time=None, raise_error=True):
        # pylint: disable=protected-access
        # try flushing all the buffered batch within given time
        _LOGGER.info("Partition: %r started flushing.", self.partition_id)
        if self._cur_batch:  # if there is batch, enqueue it to the buffer first
            self._buffered_queue.put(self._cur_batch)
            self._cur_batch = EventDataBatch(self._max_message_size_on_link, amqp_transport=self._amqp_transport)
        while self._buffered_queue.qsize() > 0:
            remaining_time = timeout_time - time.time() if timeout_time else None
            if (remaining_time and remaining_time > 0) or remaining_time is None:
                try:
                    batch = self._buffered_queue.get(block=False)
                except queue.Empty:
                    break
                self._buffered_queue.task_done()
                try:
                    _LOGGER.info("Partition %r is sending.", self.partition_id)
                    await self._producer.send(
                        batch,
                        timeout=timeout_time - time.time() if timeout_time else None,
                    )
                    _LOGGER.info(
                        "Partition %r sending %r events succeeded.",
                        self.partition_id,
                        len(batch),
                    )
                    try:
                        await self._on_success(batch._internal_events, self.partition_id)
                    except AttributeError:
                        await self._on_success(batch, self.partition_id)
                except Exception as exc:  # pylint: disable=broad-except
                    _LOGGER.info(
                        "Partition %r sending %r events failed due to exception: %r",
                        self.partition_id,
                        len(batch),
                        exc,
                    )
                    try:
                        await self._on_error(batch._internal_events, self.partition_id, exc)
                    except AttributeError:
                        await self._on_error(batch, self.partition_id, exc)
                finally:
                    self._cur_buffered_len -= len(batch)
            # If flush could not get the semaphore, we log and raise error if wanted
            else:
                _LOGGER.info(
                    "Partition %r fails to flush due to timeout.", self.partition_id
                )
                if raise_error:
                    raise OperationTimeoutError(
                        "Failed to flush {!r} within {}".format(
                            self.partition_id, timeout_time
                        )
                    )
                break
        # after finishing flushing, reset cur batch and put it into the buffer
        self._last_send_time = time.time()
        #reset curr_buffered
        self._cur_buffered_len = 0
        self._cur_batch = EventDataBatch(self._max_message_size_on_link, amqp_transport=self._amqp_transport)
        _LOGGER.info("Partition %r finished flushing.", self.partition_id)

    async def check_max_wait_time_worker(self):
        while self._running:
            if self._cur_buffered_len > 0:
                now_time = time.time()
                _LOGGER.info(
                    "Partition %r worker is checking max_wait_time.", self.partition_id
                )
                # flush the partition if its beyond the waiting time or the buffer is at max capacity
                if (now_time - self._last_send_time > self._max_wait_time) or (
                    self._cur_buffered_len >= self._max_buffer_len
                ):
                    # in the worker, not raising error for flush, users can not handle this
                    async with self._lock:
                        await self._flush(raise_error=False)
            await asyncio.sleep(min(self._max_wait_time, 5))

    @property
    def buffered_event_count(self):
        return self._cur_buffered_len



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union
from threading import Lock
from enum import Enum

from ._constants import TransportType

if TYPE_CHECKING:
    from ._pyamqp.authentication import JWTTokenAuth
    from ._pyamqp._connection import Connection
    try:
        from uamqp.authentication import JWTTokenAuth as uamqp_JWTTokenAuth
        from uamqp import Connection as uamqp_Connection
    except ImportError:
        uamqp_JWTTokenAuth = None
        uamqp_Connection = None
    from ._transport._base import AmqpTransport

    try:
        from typing_extensions import Protocol
    except ImportError:
        Protocol = object  # type: ignore

    class ConnectionManager(Protocol):
        def get_connection(
            self,
            *,
            host: Optional[str] = None,
            auth: Optional[Union[JWTTokenAuth, uamqp_JWTTokenAuth]] = None,
            endpoint: Optional[str] = None,
        ) -> Union[Connection, uamqp_Connection]:
            pass

        def close_connection(self):
            pass

        def reset_connection_if_broken(self):
            pass


class _ConnectionMode(Enum):
    ShareConnection = 1
    SeparateConnection = 2


class _SharedConnectionManager(object):  # pylint:disable=too-many-instance-attributes
    def __init__(self, **kwargs):
        self._lock = Lock()
        self._conn: Union[Connection, uamqp_Connection] = None

        self._container_id = kwargs.get("container_id")
        self._custom_endpoint_address = kwargs.get("custom_endpoint_address")
        self._debug = kwargs.get("debug")
        self._error_policy = kwargs.get("error_policy")
        self._properties = kwargs.get("properties")
        self._encoding = kwargs.get("encoding") or "UTF-8"
        self._transport_type = kwargs.get("transport_type") or TransportType.Amqp
        self._http_proxy = kwargs.get("http_proxy")
        self._max_frame_size = kwargs.get("max_frame_size")
        self._channel_max = kwargs.get("channel_max")
        self._idle_timeout = kwargs.get("idle_timeout")
        self._remote_idle_timeout_empty_frame_send_ratio = kwargs.get("remote_idle_timeout_empty_frame_send_ratio")
        self._amqp_transport: AmqpTransport = kwargs.pop("amqp_transport")

    def get_connection(
        self,
        *,
        host: Optional[str] = None,
        auth: Optional[Union[JWTTokenAuth, uamqp_JWTTokenAuth]] = None,
        endpoint: Optional[str] = None,
    ) -> Union[Connection, uamqp_Connection]:
        with self._lock:
            if self._conn is None:
                self._conn = self._amqp_transport.create_connection(
                    host=host,
                    auth=auth,
                    endpoint=endpoint,
                    custom_endpoint_address=self._custom_endpoint_address,
                    container_id=self._container_id,
                    max_frame_size=self._max_frame_size,
                    channel_max=self._channel_max,
                    idle_timeout=self._idle_timeout,
                    properties=self._properties,
                    remote_idle_timeout_empty_frame_send_ratio=self._remote_idle_timeout_empty_frame_send_ratio,
                    error_policy=self._error_policy,
                    debug=self._debug,
                    encoding=self._encoding,
                )
            return self._conn

    def close_connection(self):
        # type: () -> None
        with self._lock:
            if self._conn:
                self._amqp_transport.close_connection(self._conn)
            self._conn = None

    def reset_connection_if_broken(self):
        # type: () -> None
        with self._lock:
            conn_state = self._amqp_transport.get_connection_state(self._conn)
            if self._conn and conn_state in self._amqp_transport.CONNECTION_CLOSING_STATES:
                self._conn = None


class _SeparateConnectionManager(object):
    def __init__(self, **kwargs):
        pass

    def get_connection(  # pylint:disable=unused-argument
        self,
        *,
        host: Optional[str] = None,
        auth: Optional[Union[JWTTokenAuth, uamqp_JWTTokenAuth]] = None,
        endpoint: Optional[str] = None,
    ) -> None:
        return None

    def close_connection(self):
        # type: () -> None
        pass

    def reset_connection_if_broken(self):
        # type: () -> None
        pass


def get_connection_manager(**kwargs):
    # type: (...) -> 'ConnectionManager'
    connection_mode = kwargs.get("connection_mode", _ConnectionMode.SeparateConnection)  # type: ignore
    if connection_mode == _ConnectionMode.ShareConnection:
        return _SharedConnectionManager(**kwargs)
    return _SeparateConnectionManager(**kwargs)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events with different options to an Event Hub asynchronously.
"""

import time
import asyncio
import os

from azure.eventhub.aio import EventHubProducerClient
from azure.eventhub.exceptions import EventHubError
from azure.eventhub import EventData

CONNECTION_STR = os.environ['EVENT_HUB_CONN_STR']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']


async def send_event_data_batch(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch = await producer.create_batch()
    event_data_batch.add(EventData('Single message'))
    await producer.send_batch(event_data_batch)


async def send_event_data_batch_with_limited_size(producer):
    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.
    event_data_batch_with_limited_size = await producer.create_batch(max_size_in_bytes=1000)

    while True:
        try:
            event_data_batch_with_limited_size.add(EventData('Message inside EventBatchData'))
        except ValueError:
            # EventDataBatch object reaches max_size.
            # New EventDataBatch object can be created here to send more data.
            break

    await producer.send_batch(event_data_batch_with_limited_size)


async def send_event_data_batch_with_partition_key(producer):
    # Specifying partition_key
    event_data_batch_with_partition_key = await producer.create_batch(partition_key='pkey')
    event_data_batch_with_partition_key.add(EventData('Message will be sent to a partition determined by the partition key'))

    await producer.send_batch(event_data_batch_with_partition_key)


async def send_event_data_batch_with_partition_id(producer):
    # Specifying partition_id.
    event_data_batch_with_partition_id = await producer.create_batch(partition_id='0')
    event_data_batch_with_partition_id.add(EventData('Message will be sent to target-id partition'))

    await producer.send_batch(event_data_batch_with_partition_id)


async def send_event_data_batch_with_properties(producer):
    event_data_batch = await producer.create_batch()
    event_data = EventData('Message with properties')
    event_data.properties = {'prop_key': 'prop_value'}
    event_data_batch.add(event_data)
    await producer.send_batch(event_data_batch)


async def send_event_data_list(producer):
    # If you know beforehand that the list of events you have will not exceed the
    # size limits, you can use the `send_batch()` api directly without creating an EventDataBatch

    # Without specifying partition_id or partition_key
    # the events will be distributed to available partitions via round-robin.

    event_data_list = [EventData('Event Data {}'.format(i)) for i in range(10)]
    try:
        await producer.send_batch(event_data_list)
    except ValueError:  # Size exceeds limit. This shouldn't happen if you make sure before hand.
        print("Size of the event data list exceeds the size limit of a single send")
    except EventHubError as eh_err:
        print("Sending error: ", eh_err)


async def run():

    producer = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME
    )
    async with producer:
        await send_event_data_batch(producer)
        await send_event_data_batch_with_limited_size(producer)
        await send_event_data_batch_with_partition_key(producer)
        await send_event_data_batch_with_partition_id(producer)
        await send_event_data_batch_with_properties(producer)
        await send_event_data_list(producer)


start_time = time.time()
asyncio.run(run())
print("Send messages in {} seconds.".format(time.time() - start_time))



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# This EventHubs resource preparer is for future use after the tests are all compatible with
# The resource preparer.

import functools

from azure.mgmt.eventhub import EventHubManagementClient
from azure.mgmt.eventhub.models import Eventhub, AccessRights

from azure_devtools.scenario_tests.exceptions import AzureTestError

from devtools_testutils import (
    ResourceGroupPreparer, AzureMgmtPreparer, FakeResource
)

from devtools_testutils.resource_testcase import RESOURCE_GROUP_PARAM

EVENTHUB_DEFAULT_AUTH_RULE_NAME = 'RootManageSharedAccessKey'
EVENTHUB_NAMESPACE_PARAM = 'eventhub_namespace'
EVENTHUB_PARAM = 'eventhub'
EVENTHUB_AUTHORIZATION_RULE_PARAM = 'eventhub_authorization_rule'

# Service Bus Namespace Preparer and its shorthand decorator
class EventHubNamespacePreparer(AzureMgmtPreparer):
    def __init__(self,
                 name_prefix='',
                 use_cache=False,
                 sku='Standard', location='westus',
                 parameter_name=EVENTHUB_NAMESPACE_PARAM,
                 resource_group_parameter_name=RESOURCE_GROUP_PARAM,
                 disable_recording=True, playback_fake_resource=None,
                 client_kwargs=None, random_name_enabled=True):
        super(EventHubNamespacePreparer, self).__init__(name_prefix, 24,
                                                          random_name_enabled=random_name_enabled,
                                                          disable_recording=disable_recording,
                                                          playback_fake_resource=playback_fake_resource,
                                                          client_kwargs=client_kwargs)
        self.location = location
        self.sku = sku
        self.resource_group_parameter_name = resource_group_parameter_name
        self.parameter_name = parameter_name
        self.connection_string = ''
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "ehname"

        self.set_cache(use_cache, sku, location)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)  # type: EventHubManagementClient
            group = self._get_resource_group(**kwargs)
            namespace_async_operation = self.client.namespaces.begin_create_or_update(
                group.name,
                name,
                {
                    'sku': {'name': self.sku},
                    'location': self.location,
                }
            )
            self.resource = namespace_async_operation.result()

            key = self.client.namespaces.list_keys(group.name, name, EVENTHUB_DEFAULT_AUTH_RULE_NAME)
            self.connection_string = key.primary_connection_string
            self.key_name = key.key_name
            self.primary_key = key.primary_key
        else:
            self.resource = FakeResource(name=name, id=name)
            self.connection_string = 'Endpoint=sb://{}.servicebus.windows.net/;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='.format(name)
            self.key_name = EVENTHUB_DEFAULT_AUTH_RULE_NAME
            self.primary_key = 'ZmFrZV9hY29jdW50X2tleQ=='
        return {
            self.parameter_name: self.resource,
            '{}_connection_string'.format(self.parameter_name): self.connection_string,
            '{}_key_name'.format(self.parameter_name): self.key_name,
            '{}_primary_key'.format(self.parameter_name): self.primary_key,
            '{}_management_client'.format(self.parameter_name): self.client
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            self.client.namespaces.begin_delete(group.name, name).wait()

    def _get_resource_group(self, **kwargs):
        try:
            return kwargs.get(self.resource_group_parameter_name)
        except KeyError:
            template = 'To create an event hub a resource group is required. Please add ' \
                       'decorator @{} in front of this event hub preparer.'
            raise AzureTestError(template.format(ResourceGroupPreparer.__name__))


# Shared base class for event hub sub-resources that require a namespace and RG to exist.
class _EventHubChildResourcePreparer(AzureMgmtPreparer):
    def __init__(self,
                 name_prefix='',
                 resource_group_parameter_name=RESOURCE_GROUP_PARAM,
                 eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
                 disable_recording=True, playback_fake_resource=None,
                 client_kwargs=None, random_name_enabled=True):
        super(_EventHubChildResourcePreparer, self).__init__(name_prefix, 24,
                                                               random_name_enabled=random_name_enabled,
                                                               disable_recording=disable_recording,
                                                               playback_fake_resource=playback_fake_resource,
                                                               client_kwargs=client_kwargs)
        self.resource_group_parameter_name = resource_group_parameter_name
        self.eventhub_namespace_parameter_name = eventhub_namespace_parameter_name

    def _get_resource_group(self, **kwargs):
        try:
            return kwargs.get(self.resource_group_parameter_name)
        except KeyError:
            template = 'To create this event hub child resource event hub a resource group is required. Please add ' \
                       'decorator @{} in front of this event hub preparer.'
            raise AzureTestError(template.format(ResourceGroupPreparer.__name__))

    def _get_namespace(self, **kwargs):
        try:
            return kwargs.get(self.eventhub_namespace_parameter_name)
        except KeyError:
            template = 'To create this event hub child resource a event hub namespace is required. Please add ' \
                       'decorator @{} in front of this event hub preparer.'
            raise AzureTestError(template.format(EventHubNamespacePreparer.__name__))


class EventHubPreparer(_EventHubChildResourcePreparer):
    def __init__(self,
                 name_prefix='',
                 use_cache=False,
                 parameter_name=EVENTHUB_PARAM,
                 resource_group_parameter_name=RESOURCE_GROUP_PARAM,
                 eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
                 disable_recording=True, playback_fake_resource=None,
                 client_kwargs=None, random_name_enabled=True):
        super(EventHubPreparer, self).__init__(name_prefix,
                                                     random_name_enabled=random_name_enabled,
                                                     resource_group_parameter_name=resource_group_parameter_name,
                                                     eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
                                                     disable_recording=disable_recording,
                                                     playback_fake_resource=playback_fake_resource,
                                                     client_kwargs=client_kwargs)
        self.parameter_name = parameter_name
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "eventhub"
        self.set_cache(use_cache)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.resource = self.client.event_hubs.create_or_update(
                group.name,
                namespace.name,
                name,
                {}
            )
        else:
            self.resource = FakeResource(name=name, id=name)
        return {
            self.parameter_name: self.resource,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.client.event_hubs.delete(group.name, namespace.name, name, polling=False)


class EventHubNamespaceAuthorizationRulePreparer(_EventHubChildResourcePreparer):
    def __init__(self,
                 name_prefix='',
                 use_cache=False,
                 access_rights=[AccessRights.manage, AccessRights.send, AccessRights.listen],
                 parameter_name=EVENTHUB_AUTHORIZATION_RULE_PARAM,
                 resource_group_parameter_name=RESOURCE_GROUP_PARAM,
                 eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
                 disable_recording=True, playback_fake_resource=None,
                 client_kwargs=None, random_name_enabled=True):
        super(EventHubNamespaceAuthorizationRulePreparer, self).__init__(name_prefix,
                                                     random_name_enabled=random_name_enabled,
                                                     resource_group_parameter_name=resource_group_parameter_name,
                                                     eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
                                                     disable_recording=disable_recording,
                                                     playback_fake_resource=playback_fake_resource,
                                                     client_kwargs=client_kwargs)
        self.parameter_name = parameter_name
        self.access_rights = access_rights
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "sbnameauth"
        self.set_cache(use_cache, access_rights)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.resource = self.client.namespaces.create_or_update_authorization_rule(
                group.name,
                namespace.name,
                name,
                self.access_rights
            )

            key = self.client.namespaces.list_keys(group.name, namespace.name, name)
            connection_string = key.primary_connection_string
        else:
            self.resource = FakeResource(name=name, id=name)
            connection_string = 'https://microsoft.com'
        return {
            self.parameter_name: self.resource,
            '{}_connection_string'.format(self.parameter_name): connection_string,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            self.client.namespaces.delete_authorization_rule(group.name, namespace.name, name, polling=False)


class EventHubAuthorizationRulePreparer(_EventHubChildResourcePreparer):
    def __init__(self,
                 name_prefix='',
                 use_cache=False,
                 access_rights=[AccessRights.manage, AccessRights.send, AccessRights.listen],
                 parameter_name=EVENTHUB_AUTHORIZATION_RULE_PARAM,
                 resource_group_parameter_name=RESOURCE_GROUP_PARAM,
                 eventhub_namespace_parameter_name=EVENTHUB_NAMESPACE_PARAM,
                 eventhub_parameter_name=EVENTHUB_PARAM,
                 disable_recording=True, playback_fake_resource=None,
                 client_kwargs=None, random_name_enabled=True):
        super(EventHubAuthorizationRulePreparer, self).__init__(name_prefix,
                                                     random_name_enabled=random_name_enabled,
                                                     resource_group_parameter_name=resource_group_parameter_name,
                                                     eventhub_namespace_parameter_name=eventhub_namespace_parameter_name,
                                                     disable_recording=disable_recording,
                                                     playback_fake_resource=playback_fake_resource,
                                                     client_kwargs=client_kwargs)
        self.parameter_name = parameter_name
        self.access_rights = access_rights
        self.eventhub_parameter_name = eventhub_parameter_name
        if random_name_enabled:
            self.resource_moniker = self.name_prefix + "evnethubauth"
        self.set_cache(use_cache, access_rights)

    def create_resource(self, name, **kwargs):
        if self.is_live:
            self.client = self.create_mgmt_client(EventHubManagementClient)
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            eventhub = self._get_eventhub(**kwargs)
            self.resource = self.client.event_hubs.create_or_update_authorization_rule(
                group.name,
                namespace.name,
                eventhub.name,
                name,
                self.access_rights
            )

            key = self.client.event_hubs.list_keys(group.name, namespace.name, eventhub.name, name)
            connection_string = key.primary_connection_string
        else:
            self.resource = FakeResource(name=name, id=name)
            connection_string = 'https://microsoft.com'
        return {
            self.parameter_name: self.resource,
            '{}_connection_string'.format(self.parameter_name): connection_string,
        }

    def remove_resource(self, name, **kwargs):
        if self.is_live:
            group = self._get_resource_group(**kwargs)
            namespace = self._get_namespace(**kwargs)
            eventhub = self._get_eventhub(**kwargs)
            self.client.event_hubs.delete_authorization_rule(group.name, namespace.name, eventhub.name, name, polling=False)

    def _get_eventhub(self, **kwargs):
        try:
            return kwargs.get(self.eventhub_parameter_name)
        except KeyError:
            template = 'To create this event hub authorization rule a event hub is required. Please add ' \
                       'decorator @{} in front of this event hub preparer.'
            raise AzureTestError(template.format(EventHubPreparer.__name__))


CachedEventHubNamespacePreparer = functools.partial(EventHubNamespacePreparer, use_cache=True)
CachedEventHubPreparer = functools.partial(EventHubPreparer, use_cache=True)
